# Copyright (c) 2020-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Operator for interfacing with the Maximum Likelihood Mapmaker.

"""

import os

import numpy as np
import traitlets
from astropy import units as u
from pixell import enmap, tilemap, fft, utils, bunch

import toast
from toast.traits import trait_docs, Unicode, Int, Instance, Bool, Float, List
from toast.ops import Operator
from toast.utils import Logger, Environment, rate_from_times
from toast.timing import function_timer, Timer
from toast.observation import default_values as defaults
from toast.fft import FFTPlanReal1DStore
from toast.instrument_coords import xieta_to_quat, quat_to_xieta

import so3g

# The mapmaking tools import pixell.mpi, which will try to import
# mpi4py if it is available, even if running on a login node.
# Check to see if MPI is enabled in toast and if not, disable here.
if not toast.mpi.use_mpi:
    os.environ["DISABLE_MPI"] = "true"

from ... import mapmaking as mm
from ...core import AxisManager, IndexAxis, OffsetAxis, LabelAxis, FlagManager


@trait_docs
class MLMapmaker(Operator):
    """Operator which accumulates data to the Maximum Likelihood Mapmaker.

    """

    # Class traits

    # SKN note: I think this style is *much* more readable, but I won't change it for now
    #API       = Int(0, help="Internal interface version for this operator")
    #out_dir   = Unicode(".", help="The output directory")
    #area      = Unicode(None, allow_none=True, help="Load the enmap geometry from this file")
    #center_at = Unicode(None, allow_none=True, help="The format is [from=](ra:dec|name),[to=(ra:dec|name)],[up=(ra:dec|name|system)]")
    #comps     = Unicode("T", help="Components (must be 'T', 'QU' or 'TQU')")
    #Nmat      = Instance(allow_none=True, klass=mm.Nmat, help="The noise matrix to use")
    #dtype_map = Instance(klass=np.dtype, args=(np.float64,), help="Numpy dtype of map products")
    #times     = Unicode(obs_names.times, help="Observation shared key for timestamps")
    #boresight = Unicode(obs_names.boresight_azel, help="Observation shared key for boresight Az/El")
    #det_data  = Unicode(obs_names.det_data, help="Observation detdata key for the timestream data")
    #det_flags = Unicode(None, allow_none=True, help="Observation detdata key for flags to use")
    #det_flag_mask = Int(0, help="Bit mask value for optional detector flagging")
    #shared_flags  = Unicode(None, allow_none=True, help="Observation shared key for telescope flags to use")
    #shared_flag_mask = Int(0, help="Bit mask value for optional shared flagging")
    #view      = Unicode(None, allow_none=True, help="Use this view of the data in all observations")
    #noise_model  = Unicode("noise_model", help="Observation key containing the noise model")
    #purge_det_data = Bool(False, help="If True, clear all observation detector data after accumulating")
    #tiled     = Bool(False, help="If True, the map will be represented as distributed tiles in memory. For large maps this is faster and more memory efficient, but for small maps it has some overhead due to extra communication.")
    #verbose   = Int(1, allow_none=True, help="Set verbosity in MLMapmaker.  If None, use toast loglevel")
    #weather   = Unicode("typical", help="Weather to assume when making maps")
    #site      = Unicode("so",      help="Site to use when making maps")


    API = Int(0, help="Internal interface version for this operator")

    out_dir = Unicode(".", help="The output directory")

    area = Unicode(None, allow_none=True, help="Load the enmap geometry from this file")

    center_at = Unicode(
        None,
        allow_none=True,
        help="The format is [from=](ra:dec|name),[to=(ra:dec|name)],[up=(ra:dec|name|system)]",
    )

    interpol = Unicode(
        "nearest",
        help="Either one or a comma-separated list of interpolation modes",
    )

    downsample = List(
        [1],
        help="Downsample TOD by these factors.",
    )

    comps = Unicode("TQU", help="Components (must be 'T', 'QU' or 'TQU')")

    Nmat = Instance(klass=mm.Nmat, allow_none=True, help="The noise matrix to use")

    nmat_type = Unicode(
        "NmatDetvecs",
        help="Noise matrix type is either `NmatDetvecs`, `NmatUncorr`, `NmatWhite`, `NmatUnit` or `Nmat`",
    )

    downweight = List(
        None,
        allow_none=True,
        help="Downweight the lowest frequency bins when using NmatDetvecs noise model."
        "Set to empty list [] or None would disable downweighting."
        "When using NmatDetvecs noise model, the noise covariance N = D + VV'"
        "is the sum of uncorrelated part and auto-power in each frequency bin."
        "ACT uses downweight = [1e-4, 0.25, 0.50], the lowest three frequency bins in D "
        "will multiply by factor [1e4, 4.0, 2.0]."
        "This option is ignored when using other noise models.",
    )

    nmat_mode = Unicode(
        "build",
        help="How to initialize the noise matrix.  "
        "'build': Always build from data in obs.  "
        "'cache': Use if available in nmat_dir, otherwise build and save.  "
        "'load': Load from nmat_dir, error if missing.  "
        "'save': Build from obs data and save."
    )

    nmat_dir = Unicode(
        None,
        allow_none=True,
        help="Where to read/write/cache noise matrices. See nmat_mode. "
        "If None, write to {out_dir}/nmats"
    )

    srcsamp = Unicode(
        None,
        allow_none=True,
        help="path to mask file where True regions indicate where bright object "
        "mitigation should be applied. Mask is in equatorial coordinates. "
        "Not tiled, so should be low-res to not waste memory."
    )

    dtype_map = Unicode("float64", help="Numpy dtype of map products")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    boresight = Unicode(
        defaults.boresight_azel, help="Observation shared key for boresight Az/El"
    )

    det_data = Unicode(
        defaults.det_data, help="Observation detdata key for the timestream data"
    )

    det_flags = Unicode(
        None, allow_none=True, help="Observation detdata key for flags to use"
    )

    det_flag_mask = Int(
        defaults.det_mask_invalid, help="Bit mask value for optional detector flagging"
    )

    shared_flags = Unicode(
        None, allow_none=True, help="Observation shared key for telescope flags to use"
    )

    shared_flag_mask = Int(
        defaults.shared_mask_invalid, help="Bit mask value for optional shared flagging"
    )

    view = Unicode(
        None, allow_none=True, help="Use this view of the data in all observations"
    )

    noise_model = Unicode(
        "noise_model", help="Observation key containing the noise model"
    )

    purge_det_data = Bool(
        False,
        help="If True, clear all observation detector data after accumulating",
    )

    checkpoint_interval = Int(
        0,
        help="If greater than zero, the CG solver will store its state and"
        "restart from a checkpoint when available.",
    )

    skip_existing = Bool(
        False,
        help="If True, the mapmaker will not write any map products that "
        "already exist on disk.  See `checkpoint`."
    )

    tiled = Bool(
        True,
        help="If True, the map will be represented as distributed tiles in memory. "
        "For large maps this is faster and more memory efficient, but for small "
        "maps it has some overhead due to extra communication."
    )

    deslope = Bool(
        True,
        help="If True, each observation will have the mean and slope removed.",
    )

    verbose = Int(
        1,
        allow_none=True,
        help="Set verbosity in MLMapmaker.  If None, use toast loglevel",
    )

    weather = Unicode("vacuum", help="Weather to assume when making maps")
    site    = Unicode("so",     help="Site to use when making maps")

    maxiter = List(
        [500],
        help="List of maximum number of CG iterations for each pass.",
    )

    maxerr = Float(1e-6, help="Maximum error in the CG solver")

    truncate_tod = Bool(
        True,
        help="Truncate TOD to an easily factorizable length to ensure efficient FFT.",
    )

    write_div = Unicode(
        "all",
        allow_none=True,
        help="Components (must be 'T', 'QU', 'TQU', 'all', None, or '')"
    )

    write_hits= Bool(True, help="Write out the hitcount map")

    write_rhs = Bool(
        True, help="Write out the right hand side of the mapmaking equation"
    )

    write_bin = Bool(True, help="Write out the binned map")

    write_iter_map = Int(
        10, help="Number of iterations between saved maps.  Set to zero to disable."
    )

    @traitlets.validate("comps")
    def _check_comps(self, proposal):
        check = proposal["value"]
        if check not in ["T", "QU", "TQU"]:
            raise traitlets.TraitError("Invalid comps (must be 'T', 'QU' or 'TQU')")
        return check

    @traitlets.validate("checkpoint_interval")
    def _check_checkpoint_interval(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Invalid checkpoint_interval. Must be non-negative.")
        return check

    @traitlets.validate("shared_flag_mask")
    def _check_shared_flag_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Shared flag mask should be a positive integer")
        return check

    @traitlets.validate("det_flag_mask")
    def _check_det_flag_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Det flag mask should be a positive integer")
        return check

    @traitlets.validate("dtype_map")
    def _check_dtype_map(self, proposal):
        check = proposal["value"]
        if check not in ["float", "float64"]:
            raise traitlets.TraitError(
                "Map data type must be float64 until so3g.ProjEng supports "
                "other map data types."
            )
        return check

    @traitlets.validate("verbose")
    def _check_params(self, proposal):
        check = proposal["value"]
        if check is None:
            # Set verbosity from the toast loglevel
            env = Environment.get()
            level = env.log_level()
            if level == "VERBOSE":
                check = 3
            elif level == "DEBUG":
                check = 2
            else:
                check = 1
        return check

    @traitlets.validate("maxerr")
    def _check_maxerr(self, proposal):
        check = proposal["value"]
        if check <= 0:
            raise traitlets.TraitError("Maxerr should be greater than zero")
        return check

    @traitlets.validate("write_iter_map")
    def _check_maxerr(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("write_iter_map must be nonnegative")
        return check

    @traitlets.validate("nmat_type")
    def _check_nmat_type(self, proposal):
        check = proposal["value"]
        allowed = ["NmatUncorr", "NmatDetvecs", "NmatWhite", "NmatUnit", "Nmat"]
        if check not in allowed:
            msg = f"nmat_type must be one of {allowed}, not {check}"
            raise traitlets.TraitError(msg)
        return check

    @traitlets.validate("write_div")
    def _check_write_div(self, proposal):
        check = proposal["value"]
        if check not in ["T", "QU", "TQU", None, 'all', '']:
            raise traitlets.TraitError("Invalid write_div (must be 'T', 'QU', 'TQU', 'all', None or '')")
        return check

    def __init__(self, **kwargs):
        self.shape = None
        self.wcs = None
        self.recenter = None
        self.signal_map = None
        self.mapmaker = None
        super().__init__(**kwargs)

    @function_timer
    def setup_passes(self):
        tmp = bunch.Bunch()
        tmp.downsample = self.downsample
        tmp.maxiter = self.maxiter
        tmp.interpol = self.interpol.split(",")
        # The entries may have different lengths. We use the max
        # and then pad the others by repeating the last element.
        # The final output will be a list of bunches
        npass = max([len(tmp[key]) for key in tmp])
        passes = []
        for i in range(npass):
            entry = bunch.Bunch()
            for key in tmp:
                entry[key] = tmp[key][min(i, len(tmp[key]) - 1)]
            passes.append(entry)
        return passes

    @function_timer
    def _load_noise_model(self, ob, npass, ipass, gcomm):
        # Maybe load precomputed noise model
        log = Logger.get()
        if self.nmat_dir is None:
            nmat_dir  = os.path.join(self.out_dir, "nmats")
        else:
            nmat_dir  = self.nmat_dir
        if npass != 1:
            nmat_dir += f"_pass{ipass + 1}"
        nmat_file = os.path.join(nmat_dir, f"nmat_{ob.name}.hdf")
        there = os.path.isfile(nmat_file)
        if self.nmat_mode == "load" and not there:
            raise RuntimeError(
                f"Nmat mode is 'load' but {nmat_file} does not exist."
            )
        if self.nmat_mode == "load" or (self.nmat_mode == "cache" and there):
            log.debug_rank(f"Loading noise model from '{nmat_file}'", comm=gcomm)
            try:
                nmat = mm.read_nmat(nmat_file)
            except Exception as e:
                if self.nmat_mode == "cache":
                    log.info_rank(
                        f"Failed to load noise model from '{nmat_file}'"
                        f" : '{e}'. Will cache a new one",
                        comm=gcomm,
                    )
                    nmat = None
                else:
                    msg = f"Failed to load noise model from '{nmat_file}' : {e}"
                    raise RuntimeError(msg)
        else:
            nmat = None
        return nmat, nmat_file

    @function_timer
    def _save_noise_model(self, mapmaker, nmat, nmat_file, gcomm):
        # Maybe save the noise model we built (only if we actually built one rather than
        # reading one in)
        log = Logger.get()
        if self.nmat_mode in ["save", "cache"] and nmat is None:
            log.debug_rank(f"Writing noise model to '{nmat_file}'", comm=gcomm)
            nmat_dir = os.path.dirname(nmat_file)
            os.makedirs(nmat_dir, exist_ok=True)
            mm.write_nmat(nmat_file, mapmaker.data[-1].nmat)
        return

    @function_timer
    def _wrap_obs(self, ob, dets, passinfo):
        """ Prepare data for the mapmaker """

        # Get the focalplane for this observation
        fp = ob.telescope.focalplane

        # Get the sample rate from the data.  We also have nominal sample rates
        # from the noise model and also from the focalplane.
        # (rate, dt, dt_min, dt_max, dt_std) = rate_from_times(
        #     ob.shared[self.times].data
        # )

        axdets = LabelAxis("dets", dets)
        nsample = int(ob.n_local_samples)

        axsamps = OffsetAxis(
            "samps",
            count=nsample,
            offset=ob.local_index_offset,
            origin_tag=ob.name,
        )

        # Convert the data view into a RangesMatrix
        ranges = so3g.proj.ranges.RangesMatrix.zeros((len(dets), nsample))
        if self.view is not None:
             view_ranges = np.array(
                 [[x.first, x.last] for x in ob.intervals[self.view]]
             )
             ranges += so3g.proj.ranges.Ranges.from_array(view_ranges, nsample)

        # Convert the focalplane offsets into the expected form
        det_to_row = {y["name"]: x for x, y in enumerate(fp.detector_data)}
        det_quat = np.array([fp.detector_data["quat"][det_to_row[x]] for x in dets])
        xi, eta, gamma = quat_to_xieta(det_quat)

        axfp = AxisManager()
        axfp.wrap("xi", xi, axis_map=[(0, axdets)])
        axfp.wrap("eta", eta, axis_map=[(0, axdets)])
        axfp.wrap("gamma", gamma, axis_map=[(0, axdets)])

        # Convert Az/El quaternion of the detector back into
        # angles from the simulation.
        theta, phi, pa = toast.qarray.to_iso_angles(ob.shared[self.boresight])

        # Azimuth is measured in the opposite direction from longitude
        az = 2 * np.pi - phi
        el = np.pi / 2 - theta
        roll = pa

        axbore = AxisManager()
        axbore.wrap("az", az, axis_map=[(0, axsamps)])
        axbore.wrap("el", el, axis_map=[(0, axsamps)])
        axbore.wrap("roll", roll, axis_map=[(0, axsamps)])

        axobs = AxisManager()
        axobs.wrap("focal_plane", axfp)
        axobs.wrap("timestamps", ob.shared[self.times], axis_map=[(0, axsamps)])
        axobs.wrap(
            "signal",
            np.vstack(ob.detdata[self.det_data][dets, :]),
            axis_map=[(0, axdets), (1, axsamps)],
        )
        axobs.wrap("boresight", axbore)
        axobs.wrap('flags', FlagManager.for_tod(axobs))
        axobs.flags.wrap("glitch_flags", ranges, axis_map=[(0, axdets), (1, axsamps)])
        axobs.wrap("weather", np.full(1, self.weather))
        axobs.wrap("site",    np.full(1, "so"))

        if self.truncate_tod:
            # FFT-truncate for faster fft ops
            axobs.restrict("samps", [0, fft.fft_len(axobs.samps.count)])

        # MLMapmaker.add_obs will apply deslope
        if self.deslope:
            utils.deslope(axobs.signal, w=5, inplace=True)

        if self.downsample != 1:
            axobs = mm.downsample_obs(axobs, passinfo.downsample)

        # NOTE:  Expected contents look like:
        # >>> tod
        # AxisManager(signal[dets,samps], timestamps[samps], readout_filter_cal[dets],
        # mce_filter_params[6], iir_params[3,5], flags*[samps], boresight*[samps],
        # array_data*[dets], pointofs*[dets], focal_plane*[dets], abscal[dets],
        # timeconst[dets], glitch_flags[dets,samps], source_flags[dets,samps],
        # relcal[dets], dets:LabelAxis(63), samps:OffsetAxis(372680))
        # >>> tod.focal_plane
        # AxisManager(xi[dets], eta[dets], gamma[dets], dets:LabelAxis(63))
        # >>> tod.boresight
        # AxisManager(az[samps], el[samps], roll[samps], samps:OffsetAxis(372680))

        return axobs

    @function_timer
    def _init_mapmaker(
            self, mapmaker, signal_map, mapmaker_prev, eval_prev, comm, gcomm, prefix,
    ):
        """
        This function is run at the end of the pass, runs the prepare, writes the maps
        and sets up the initial condition
        """
        log = Logger.get()
        timer = Timer()
        timer.start()

        mapmaker.prepare()

        if self.tiled:
            # Each group reports how many tiles they are using
            geo_work = mapmaker.signals[1].geo_work
            nactive = len(geo_work.active)
            ntile = np.prod(geo_work.shape[-2:])
            log.debug_rank(f"{nactive} / {ntile} tiles active", comm=gcomm)

        if comm is not None:
            comm.barrier()
        log.info_rank(
            f"MLMapmaker finished prepare in",
            comm=comm,
            timer=timer,
        )

        # This will need to be modified for more general cases where we don't solve for
        # a sky map, or where we solve for multiple sky maps. The mapmaker itself supports it,
        # the problem is the direct access to the rhs, div and idiv members
        if self.write_rhs:
            fname = f"{prefix}sky_rhs.fits"
            if self.skip_existing and os.path.isfile(fname):
                log.info_rank(f"Skipping existing rhs in {fname}", comm=comm)
            else:
                fname = signal_map.write(prefix, "rhs", signal_map.rhs)
                log.info_rank(f"Wrote rhs to {fname}", comm=comm)

        if self.write_div is not None:
            # Write each covariance element seperately, to reduce peak memory.
            for i,ci in enumerate(self.comps):
                for j,cj in enumerate(self.comps):
                    if ci in self.write_div and cj in self.write_div:
                        fname = f"{prefix}sky_div{ci}{cj}.fits"
                        if self.skip_existing and os.path.isfile(fname):
                            log.info_rank(f"Skipping existing div{ci}{cj} in {fname}", comm=comm)
                        else:
                            fname = signal_map.write(prefix, f"div{ci}{cj}", signal_map.div[i, j])
                            log.info_rank(f"Wrote div{ci}{cj} to {fname}", comm=comm)

        if self.write_hits:
            fname = f"{prefix}sky_hits.fits"
            if self.skip_existing and os.path.isfile(fname):
                log.info_rank(f"Skipping existing hits in {fname}", comm=comm)
            else:
                fname = signal_map.write(prefix, "hits", signal_map.hits)
                log.info_rank(f"Wrote hits to {fname}", comm=comm)

        mmul = tilemap.map_mul if self.tiled else enmap.map_mul
        if self.write_bin:
            fname = f"{prefix}sky_bin.fits"
            if self.skip_existing and os.path.isfile(fname):
                log.info_rank(f"Skipping existing bin in {fname}", comm=comm)
            else:
                fname = signal_map.write(
                    prefix, "bin", mmul(signal_map.idiv, signal_map.rhs)
                )
                log.info_rank(f"Wrote bin to {fname}", comm=comm)

        if comm is not None:
            comm.barrier()
        log.info_rank(f"MLMapmaker finished writing rhs, div, bin in", comm=comm, timer=timer)

        # Set up initial condition
        if eval_prev is None:
            # this will be the first pass
            x0 = None
        else:
            x0 = mapmaker.translate(mapmaker_prev, eval_prev.x_zip)
        return x0

    @function_timer
    def _apply_mapmaker(self, mapmaker, x0, passinfo, prefix, comm):
        log = Logger.get()
        timer = Timer()
        timer.start()
        tstep = Timer()
        tstep.start()

        if self.checkpoint_interval > 0:
            fname_checkpoint = f"{prefix}checkpoint.{comm.rank:04}.hdf"
            there = os.path.isfile(fname_checkpoint)
            if there:
                log.info_rank(f"Checkpoint detected. Will start from previous solver state", comm=comm)
        else:
            fname_checkpoint = None

        for step in mapmaker.solve(
                maxiter=passinfo.maxiter,
                maxerr=self.maxerr,
                x0=x0,
                fname_checkpoint=fname_checkpoint,
                checkpoint_interval=self.checkpoint_interval,
        ):
            if self.write_iter_map < 1:
                dump = False
            else:
                dump = step.i % self.write_iter_map == 0
            msg = f"CG step {step.i:4d} {step.err:15.7e} write={dump}"
            log.info_rank(f"MLMapmaker   {msg} ", comm=comm, timer=tstep)
            if dump:
                for signal, val in zip(mapmaker.signals, step.x):
                    if signal.output:
                        fname = signal.write(prefix, f"map{step.i:04}", val)
                        log.info_rank(f"Wrote signal to {fname} in", comm=comm, timer=tstep)

        log.info_rank(f"MLMapmaker finished solve in", comm=comm, timer=timer)

        for signal, val in zip(mapmaker.signals, step.x):
            if signal.output:
                fname = signal.write(prefix, "map", val)
                log.info_rank(f"Wrote {fname}", comm=comm)

        if comm is not None:
            comm.barrier()
        log.info_rank(f"MLMapmaker wrote map in", comm=comm, timer=timer)

        return mapmaker, mapmaker.evaluator(step.x_zip)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        comm = data.comm.comm_world
        gcomm = data.comm.comm_group
        timer.start()

        if self.write_div == 'all':
            self.write_div = self.comps
        elif self.write_div == '':
            self.write_div = None
        elif self.write_div is not None:
            # Make sure all components in self.write_div is in self.comps
            for i in self.write_div:
                if i not in self.comps:
                    msg = f"Component '{i}' in write_div={self.write_div} not present in comps={self.comps}"
                    raise RuntimeError(msg)

        if data.comm.group_size != 1:
            raise RuntimeError(
                "The ML mapmaker requires the TOAST process group size to be exactly one."
            )

        if comm is not None:
            num_obs = np.array(comm.allgather(len(data.obs)))
        else:
            num_obs = np.array([len(data.obs)])
        if np.any(num_obs == 0):
            # At least one TOAST group have no observation
            if np.sum(num_obs) < data.comm.ngroups:
                msg = "There are {} TOAST groups but only {} observations. ".format(data.comm.ngroups, np.sum(num_obs))
                msg += "Therefore some TOAST groups have no observation in it. "
                msg += "Please reduce the number of TOAST groups."
                raise RuntimeError(msg)
            else:
                raise RuntimeError("Some TOAST groups have no observation.")

        if comm is None and self.tiled:
            log.info("WARNING: Tiled mapmaking not supported without MPI.")
            self.tiled = False

        for trait in ["area"]:
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(
                    f"You must set `{trait}` before running MLMapmaker"
                )

        # nmat_type is guaranteed to be a valid Nmat class
        if self.nmat_type == 'NmatDetvecs':
            noise_model = getattr(mm, self.nmat_type)(downweight=self.downweight)
        else:
            noise_model = getattr(mm, self.nmat_type)()

        shape, wcs = enmap.read_map_geometry(self.area)

        if self.center_at is None:
            recenter = None
        else:
            recenter = mm.parse_recentering(self.center_at)
        dtype_tod = np.float32
        dtype_map = np.dtype(self.dtype_map)

        prefix = os.path.join(self.out_dir, f"{self.name}_")

        passes = self.setup_passes()
        npass = len(passes)
        mapmaker_prev = None
        eval_prev = None

        if self.srcsamp is None:
            srcsamp_mask = None
        else:
            if not os.path.isfile(self.srcsamp):
                raise RuntimeError(f"Source mask does not exist: {self.srcsamp}")
            srcsamp_mask = enmap.read_map(self.srcsamp)

        for ipass, passinfo in enumerate(passes):
            # The multipass mapmaking loop
            log.info_rank(
                f"Starting pass {ipass + 1}/{npass}, maxit={passinfo.maxiter} "
                f"down={passinfo.downsample}, interp={passinfo.interpol}",
                comm=comm,
            )
            if npass == 1:
                pass_prefix = prefix
            else:
                pass_prefix = f"{prefix}pass{ipass + 1}_"

            signal_cut = mm.SignalCut(comm, dtype=dtype_tod)
            signal_map = mm.SignalMap(
                shape,
                wcs,
                comm,
                comps=self.comps,
                dtype=dtype_map,
                recenter=recenter,
                tiled=self.tiled,
                interpol=passinfo.interpol,
            )
            signals = [signal_cut, signal_map]
            if srcsamp_mask is not None:
                signal_srcsamp = mm.SignalSrcsamp(comm, srcsamp_mask, dtype=dtype_tod)
                signals.append(signal_srcsamp)
            mapmaker = mm.MLMapmaker(
                signals, noise_model=noise_model, dtype=dtype_tod, verbose=self.verbose
            )

            for ob in data.obs:
                # Get the detectors we are using locally for this observation
                dets = ob.select_local_detectors(detectors)
                if len(dets) == 0:
                    # Nothing to do for this observation
                    continue

                nmat, nmat_file = self._load_noise_model(ob, npass, ipass, gcomm)

                # wrap_obs finishes in line 250 of make_ml_map.py, at the downsampling
                axobs = self._wrap_obs(ob, dets, passinfo)

                if ipass > 0:
                    # Evaluate the final model of the previous pass' mapmaker
                    # for this observation.
                    signal_estimate = eval_prev.evaluate(mapmaker_prev.data[len(mapmaker.data)])
                    # Resample this to the current downsampling level
                    signal_estimate = mm.resample.resample_fft_simple(signal_estimate, axobs.samps.count)
                else:
                    signal_estimate = None

                mapmaker.add_obs(
                    ob.name,
                    axobs,
                    deslope=self.deslope,
                    noise_model=nmat,
                    signal_estimate=signal_estimate,
                )
                del axobs
                del signal_estimate

                self._save_noise_model(mapmaker, nmat, nmat_file, gcomm)

                # Optionally delete the input detector data to save memory, if
                # the calling code knows that no additional operators will be
                # used afterwards.
                if ipass == npass - 1 and self.purge_det_data:
                    del ob.detdata[self.det_data]

            if comm is not None:
                comm.barrier()
            log.info_rank(
                f"MLMapmaker wrapped observations in",
                comm=comm,
                timer=timer,
            )

            # _init_mapmaker covers lines 293-303 of make_ml_map.py
            x0 = self._init_mapmaker(
                mapmaker,
                signal_map,
                mapmaker_prev,
                eval_prev,
                comm,
                gcomm,
                pass_prefix,
            )
            # _apply_mapmaker covers lines 305-320 of make_ml_map.py
            mapmaker_prev, eval_prev = self._apply_mapmaker(mapmaker, x0, passinfo, pass_prefix, comm)

            # Save metadata, may get dropped later

            self.shape = shape
            self.wcs = wcs
            self.recenter = recenter
            self.signal_map = signal_map
            self.mapmaker = mapmaker

        return

    @function_timer
    def _finalize(self, data, **kwargs):
        pass

    def _requires(self):
        req = {
            "meta": [self.noise_model],
            "shared": [
                self.times,
            ],
            "detdata": [self.det_data],
            "intervals": list(),
        }
        if self.view is not None:
            req["intervals"].append(self.view)
        if self.shared_flags is not None:
            req["shared"].append(self.shared_flags)
        if self.det_flags is not None:
            req["detdata"].append(self.det_flags)
        return req

    def _provides(self):
        prov = {"detdata": list()}
        if self.det_out is not None:
            prov["detdata"].append(self.det_out)
        return prov

    def _accelerators(self):
        return list()



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory Data Processing.

This module contains the core data I/O and processing tools.

"""

from so3g.spt3g import core


class DataG3Module(object):
    """
    Base G3Module for processing or conditioning data.
    
    Classes inheriting from DataG3Module only need to override the process() function
    to build a G3Module that will call process() on every detector in a given 
    G3TimestreamMap. 
    
    Attributes:
        input (str): key of G3TimestreamMap of source data
        output (str or None): key of G3Timestream map of output data.
            if None, input will be overwritten with output
            
    TODO:
        * Add detector masking capabilities so we don't waste CPU time on cut detectors
        * Make an option for input/output to be a list, so processing can be done on multiple timestreams
        * Make it possible for input/output to be a G3Timestream, instead of a G3TimestreamMap
        * Decide if "Modules that add members to frames" should be a default type
    """
    def __init__(self, input='signal', output='signal_out'):
        self.input = input
        if output is None:
            self.output = self.input
        else:
            self.output = output
                
    def __call__(self, f):
        if f.type == core.G3FrameType.Scan:
            if self.input not in f.keys() or type(f[self.input]) != core.G3TimestreamMap:
                raise ValueError("""Frame is a Scan but does not have a G3Timestream map 
                                 named {}""".format(self.input))
        
            processing = core.G3TimestreamMap()
            
            for k in f[self.input].keys():
                processing[k] = core.G3Timestream( self.process(f[self.input][k], k) )
                                    
            processing.start = f[self.input].start
            processing.stop = f[self.input].stop   

            if self.input == self.output:
                f.pop(self.input)
            f[self.output] = processing
           
    def process(self, data, det_name):
        """
        Args:
            data (G3Timestream): data for a single detector
            det_name (str): the detector name in the focal plane,
                in case it's needed for accessing calibration info
        Returns:
            data (G3Timestream)
        """
        return data
    
    def apply(self, f):
        """
        Applies the G3Module on an individual G3Frame or G3TimestreamMap. Likely
        to be useful when debugging
        
        Args:
            f (G3Frame or G3TimestreamMap)
            
        Returns:
            G3Frame: if f is a G3Frame it returns the frame after the module has been applied
            G3TimestreamMap: if f is a G3TimestreamMap it returns self.output
        """
        if type(f) == core.G3Frame:
            if f.type != core.G3FrameType.Scan:
                raise ValueError("""Cannot apply {} on a {} frame""".format(self.__class__,
                                                                           f.type))
            self.__call__(f)
            return f
        if type(f) == core.G3TimestreamMap:
            frame = core.G3Frame()
            frame.type = core.G3FrameType.Scan
            frame[self.input] = f
            self.__call__(frame)
            return frame[self.output]
        raise ValueError('apply requires a G3Frame or a G3TimestreamMap, you gave me {}'.format(type(f)))
        
    @classmethod
    def from_function(cls, function, input='signal', output=None):
        """
        Allows inline creation of DataG3Modules with just a function definition
        
        Example:
            mean_sub = lambda data: data-np.mean(data)
            G3Pipeline.Add(DataG3Module.from_function(mean_sub) )

        Args:
            function (function): function that takes a G3Timestream (or ndarray)
                and returns an ndarray
        """
        dg3m = cls(input, output)
        process = lambda self, data, det_name: function(data)
        dg3m.process = process.__get__(dg3m, DataG3Module)
        return dg3m


from .pmat import P
from .helpers import (
    get_radec, get_horiz, get_footprint, get_wcs_kernel, get_supergeom,
    Timer, DEG, ScalarLastQuat
)
from . import planets
from . import local
from .plotting import *
from . import demod
from . import pointing_model
from . import healpix_utils



"""
The purpose of ManifestDb is to provide a layer of abstraction that
assists code with identifying a file that contains some desired
information, based on a piece of index information.  For example, a
user might have a timestamp and an array name and want to know the
filename for the effective beam that should be used for that array at
that time.

**Definitions:**

Index Data
  A set of (key,value) pairs provided by the user.

Endpoint Data
  A set of (key,value) pairs returned by the Manifest.

The ManifestDb is primarily responsible for transforming Index Data
into Endpoint Data.  The implementation has the following features:

* The rules and data for mapping Index Data to Endpoint Data are
  stored in an sqlite database, for portability.

* It is assumed that a 'filename' will often be a piece of the
  Endpoint Data, so this is given a special status and table storage.

If a ManifestDb is used as part of the interface to a large metadata
archive, then minor updates to the metadata can be made by only
providing a the changed data (in a separate archive file), with an
updated ManifestDb that sends only the affected values of Index Data
to the new archive file.  When this approach is used, is suggests that
metadata archive files and the ManifestDb should be stored in slightly
different places, since there is a many-to-many relationship between
them.

When generating a ManifestDb, a ManifestScheme must first be defined.
The ManifestScheme includes:

* a description of what values may be present in the Index Data, and
  they should be used.
* a description of Endpoint Data to associate with valid Index Data.
"""

import sqlite3
import os
import sys
import json
import numpy as np
import argparse

from . import common, resultset

TABLE_DEFS = {
    'input_scheme': [
        "`id`      integer primary key autoincrement",
        "`name`    varchar(32) unique",
        "`purpose` varchar(32)",
        "`match`   varchar(32)",
        "`dtype`   varchar(32)",
        ],
    'map_template': [
        "`id` integer primary key autoincrement",
        "`file_id` integer",
    ],
    'files': [
        "`id` integer primary key autoincrement",
        "`name` varchar(256)",
    ],
}

# Because sqlite will let you store anything anywhere, the most
# sensible default column type is "numeric" -- this will automatically
# cast number-like strings to numbers (while accepting strings).
DEFAULT_DTYPE = 'numeric'


class ManifestScheme:
    def __init__(self):
        self.cols = []

    def new_db(self, **kwargs):
        """Use this scheme to instantiate a ManifestDb, and return it.  Aall
        kwargs are passed to ManifestDb constructor.

        """
        return ManifestDb(scheme=self, **kwargs)

    # Methods for constructing table.

    def add_exact_match(self, name, dtype=DEFAULT_DTYPE):
        """
        Add a field to the scheme, that must be matched exactly.
        """
        self.cols.append((name, 'in', 'exact', dtype))
        return self

    def add_range_match(self, name, purpose='in', dtype=DEFAULT_DTYPE):
        """
        Add a field to the scheme, that represents a range of values to be
        matched by a single input value.
        """
        self.cols.append((name, purpose, 'range', dtype))
        return self

    def add_data_field(self, name, dtype=DEFAULT_DTYPE):
        """
        Add a field to the scheme that is returned along with the matched
        filename.
        """
        self.cols.append((name, 'out', 'exact', dtype))
        return self

    def as_resultset(self):
        """Get the scheme structure as a ResultSet.  This is a safer
        alternative to inspecting .cols directly.

        """
        rs = resultset.ResultSet(['field', 'purpose', 'col_type', 'data_type'])
        rs.rows = list(self.cols)
        return rs

    def _get_scheme_rows(self):
        """
        Returns a list of tuples (name,purpose,match,dtype), suitable
        for populating the input_scheme table.
        """
        rows = []
        for c in self.cols:
            name, purpose, match, dtype = c
            if match == 'exact':
                rows.append((name, purpose, 'exact', dtype))
            elif match == 'range':
                rows.append((name, purpose, 'range', dtype))
            else:
                raise ValueError("Bad ctype '%s'" % match)
        return rows

    def _get_map_table_def(self):
        """
        Returns column definitions for the map table.
        """
        entries = [row for row in TABLE_DEFS['map_template']]
        uniques = []
        for c in self.cols:
            name, purpose, match, dtype = c
            if match == 'exact':
                entries.append('`%s` %s' % (name, dtype))
                uniques.append('`%s`' % name)
            elif match == 'range':
                entries.append('`%s__lo` %s' % (name, dtype))
                entries.append('`%s__hi` %s' % (name, dtype))
                uniques.append('`%s__lo`' % name)
                uniques.append('`%s__hi`' % name)
            else:
                raise ValueError("Bad ctype '%s'" % match)
        entries.append('UNIQUE(' + ','.join(uniques) + ')')
        return entries

    def _format_row(self, r):
        """Rewrite a dict of index and/or endpoint data from the database so
        that it is compatible with the format expected by
        ManifestDb.add_entry / update_entry.

        This modifies the provided object (and also returns it).

        """
        for c in self.cols:
            name, purpose, match, dtype = c
            if match == 'range':
                a, b = r.pop('%s__lo' % name), r.pop('%s__hi' % name)
                r[name] = (a, b)
        return r

    @classmethod
    def from_database(cls, conn, table_name='input_scheme'):
        """
        Decode a ManifestScheme from the provided sqlite database
        connection.
        """
        c = conn.cursor()
        c.execute('select name,purpose,match,dtype from %s' % table_name)
        self = cls()
        for r in c:
            (name, purpose, match, dtype) = r
            if match in ['exact', 'range']:
                self.cols.append((name, purpose, match, dtype))
            else:
                raise ValueError("Bad ctype '%s'" % match)
        return self

    def get_match_query(self, params, partial=False, strict=False):
        """Get sql query fragments for this ManifestDb.

        Arguments:
          params: a dict of values to match against.
          partial: if True, then operate in "inspection" mode (see notes).
          strict: if True, then reject any requests that include
            entries in params that are not known to the schema.

        Returns:
          (where_string, values_tuple, ret_cols)

        Notes:
          The normal mode (partial=False) requires that every "in"
          column in the scheme has a key=value pair in the params
          dict, and the ret_cols are the "out" columns.  In inspection
          mode (partial=True), then any column can be matched against
          the params, and the complete row data of all matching rows
          is returned.

        """
        qs = []
        vals = []
        ret_cols = []
        unassigned = dict(params)
        assert('filename' not in params)

        for col in self.cols:
            (name, purpose, match, dtype) = col
            purposes = [purpose]

            if partial:
                # in/out direction is entirely determined by whether
                # user passed a value.
                if name in params:
                    purposes = ['in', 'out']
                else:
                    purposes = ['out']
            else:
                if 'in' in purposes and name not in params:
                    raise ValueError('Parameter %s is not optional.' % name)

            if 'in' in purposes:
                if match == 'exact':
                    qs.append('`%s`=?' % name)
                    vals.append(params[name])
                elif match == 'range':
                    qs.append('(`%s__lo` <= ?) and (? < `%s__hi`)' % (name, name))
                    vals.extend([params[name], params[name]])
                else:
                    raise ValueError("Bad ctype '%s'" % match)
                unassigned.pop(name)
            if 'out' in purposes:
                # Include that column's value in result.
                if match == 'range':
                    ret_cols.extend(['%s__lo' % name, '%s__hi' % name])
                else:
                    ret_cols.append(name)
        if strict and len(unassigned):
            raise ValueError(f'Failed to match params: {unassigned}')
            assert(len(unassigned) == 0)
        return (' and '.join(qs), tuple(vals), ret_cols)

    def get_insertion_query(self, params):
        """Get sql query fragments for inserting a new entry with the provided
        params.

        Returns:
          (fields, values) where fields is a string with the field
          names (comma-delimited) and values is a tuple of values.

        """
        qs = []
        vals = []
        unassigned = list(params.keys())
        for col in self.cols:
            (name, purpose, match, dtype) = col
            if not name in params:
                raise ValueError('Parameter %s is not optional.' % name)
            unassigned.remove(name)
            if match == 'exact':
                qs.append('`%s`' % name)
                vals.append(params[name])
            elif match == 'range':
                qs.append('`%s__lo`,`%s__hi`' % (name,name))
                vals.extend(params[name])
            else:
                raise ValueError("Bad ctype '%s'" % match)
        if len(unassigned):
            raise ValueError('Attempting to add data for unknown fields: %s' % unassigned)
        return ','.join(qs), tuple(vals)

    def get_update_query(self, params):
        """Get sql query fragments for updating an entry.

        Returns:
          (setstring, values) where setstring is of the form
          "A=?,...,Z=?" and values is the corresponding tuple of values.

        """
        keys = []
        vals = []
        unassigned = [k for k in params.keys() if k not in '_id']
        for col in self.cols:
            (name, purpose, match, dtype) = col
            if not name in params:
                continue
            unassigned.remove(name)
            if match == 'exact':
                keys.append('`%s`' % name)
                vals.append(params[name])
            elif match == 'range':
                keys.extend(['`%s__lo`' % name, '`%s__hi`' % name])
                vals.extend(params[name])
            else:
                raise ValueError("Bad ctype '%s'" % match)
        if len(unassigned):
            raise ValueError('Attempting to update data for unknown fields: %s' % unassigned)
        return ','.join([f'{k}=?' for k in keys]), tuple(vals)

    def get_required_params(self):
        """
        Returns a list of parameter names that are required for matching.
        """
        return [c[0] for c in self.cols if c[1] == 'in']


class ManifestDb:
    """
    Expose a map from Index Data to Endpoint Data, including a
    filename.
    """

    def __init__(self, map_file=None, scheme=None):
        """Instantiate a database.  If map_file is provided, the
        database will be connected to the indicated sqlite file on
        disk, and any changes made to this object be written back to
        the file.

        If scheme is None, the scheme will be loaded from the
        database; pass scheme=False to prevent that and leave the db
        uninitialized.

        """
        if isinstance(map_file, sqlite3.Connection):
            self.conn = map_file
        else:
            if map_file is None:
                map_file = ':memory:'
            self.conn = sqlite3.connect(map_file)

        self.conn.row_factory = sqlite3.Row  # access columns by name

        if scheme is None:
            self.scheme = ManifestScheme.from_database(self.conn)
        elif scheme is False:
            pass
        else:
            self._create(scheme)

    def _create(self, manifest_scheme):
        """
        Create the database tables, incorporating the provided
        ManifestScheme.
        """
        # Create the tables:
        table_defs = [
            ('input_scheme', TABLE_DEFS['input_scheme']),
            ('files', TABLE_DEFS['files']),
            ('map', manifest_scheme._get_map_table_def())]
        c = self.conn.cursor()
        for table_name, column_defs in table_defs:
            q = ('create table if not exists `%s` (' % table_name  +
                 ','.join(column_defs) + ')')
            c.execute(q)
        self.conn.commit()
        # Commit the schema...
        for r in manifest_scheme._get_scheme_rows():
            c.execute('insert into input_scheme (name,purpose,match,dtype) '
                      'values (?,?,?,?)', tuple(r))
        self.conn.commit()

        self.scheme = ManifestScheme.from_database(self.conn)

    def copy(self, map_file=None, overwrite=False):
        """
        Duplicate the current database into a new database object, and
        return it.  If map_file is specified, the new database will be
        connected to that sqlite file on disk.  Note that a quick way
        of writing a Db to disk to call copy(map_file=...) and then
        simply discard the returned object.
        """
        if map_file is not None and os.path.exists(map_file):
            if overwrite:
                os.remove(map_file)
            else:
                raise RuntimeError("Output file %s exists (overwrite=True "
                                   "to overwrite)." % map_file)
        new_db = ManifestDb(map_file=map_file, scheme=False)
        script = ' '.join(self.conn.iterdump())
        new_db.conn.executescript(script)
        new_db.scheme = ManifestScheme.from_database(new_db.conn)
        return new_db

    def to_file(self, filename, overwrite=True, fmt=None):
        """Write the present database to the indicated filename.

        Args:
          filename (str): the path to the output file.
          overwrite (bool): whether an existing file should be
            overwritten.
          fmt (str): 'sqlite', 'dump', or 'gz'.  Defaults to 'sqlite'
            unless the filename ends with '.gz', in which it is 'gz'.

        """
        return common.sqlite_to_file(self.conn, filename, overwrite=overwrite, fmt=fmt)

    @classmethod
    def from_file(cls, filename, fmt=None, force_new_db=True):
        """Instantiate an ManifestDb and return it, with the data copied in
        from the specified file.

        Args:
          filename (str): path to the file.
          fmt (str): format of the input; see to_file for details.
          force_new_db (bool): Used if connecting to an sqlite
            database. If True the database is copied into memory and
            if False returns a read-only connection to the database
            without reading it into memory

        Returns:
          ManifestDb with an sqlite3 connection that is mapped to memory.

        Notes:
          Note that if you want a `persistent` connection to the file,
          you should instead pass the filename to the ManifestDb
          constructor map_file argument.

        """
        conn = common.sqlite_from_file(filename, fmt=fmt, force_new_db=force_new_db)
        return cls(conn)

    @classmethod
    def readonly(cls, filename):
        """Instantiate an ManifestDb connected to an sqlite database on disk,
        and return it.  The database remains mapped to disk, in readonly mode.

        Args:
          filename (str): path to the file.

        Returns:
          ManifestDb.

        """
        conn = sqlite3.connect('file:%s?mode=ro' % filename, uri=True)
        return cls(conn)

    def _get_file_id(self, filename, create=False):
        """
        Lookup a file_id in the file table, or create it if `create` and not found.
        """
        c = self.conn.cursor()
        c.execute('select id from files where name=?', (filename,))
        row_id = c.fetchone()
        if row_id is None:
            if create:
                c.execute('insert into files (name) values (?)', (filename,))
                row_id = c.lastrowid
                return row_id
            return None
        return row_id[0]

    def match(self, params, multi=False, prefix=None):
        """Given Index Data, return Endpoint Data.

        Arguments:
          params (dict): Index Data.
          multi (bool): Whether more than one result may be returned.
          prefix (str or None): If set, it will be os.path.join-ed to
            the filename from the db.

        Returns:
          A dict of Endpoint Data, or None if no match was found.  If
          multi=True then a list is returned, which could have 0, 1,
          or more items.

        """
        q, p, rp = self.scheme.get_match_query(params)
        cols = ['files`.`name'] + list(rp)
        c = self.conn.cursor()
        where_str = ''
        if len(q):
            where_str = 'where %s' % q
        c.execute('select `%s` ' % ('`,`'.join(cols)) + 
                  'from map join files on map.file_id=files.id %s' % where_str, p)
        rows = c.fetchall()
        rp.insert(0, 'filename')
        rows = [dict(zip(rp, r)) for r in rows]
        if prefix is not None:
            for r in rows:
                r['filename'] = os.path.join(prefix, r['filename'])
        if multi:
            return rows
        if len(rows) == 0:
            return None
        if len(rows) > 1:
            raise ValueError('Matched multiple rows with index data: %s' % rows)
        return rows[0]

    def inspect(self, params={}, strict=True, prefix=None):
        """Given (partial) Index Data and Endpoint Data, find and return the
        complete matching records.

        Arguments:
          params (dict): any mix of Index Data and Endpoint Data.
          strict (bool): if True, a ValueError will be raised if
            params contains any keys that aren't recognized as Index
            or Endpoint data.
          prefix (str or None): As in .match().

        Returns:
          A list of results matching the query.  Each result in the
          list is a dict, containing complete entry data.  A special
          entry, '_id', is the database row id and can be used to
          update or remove specific entries.

        """
        params = dict(params)
        filename = params.pop('filename', None)

        q, p, rp = self.scheme.get_match_query(params, partial=True, strict=strict)
        cols = ['map`.`id', 'files`.`name'] + list(rp)
        rp = ['_id', 'filename'] + rp
        c = self.conn.cursor()
        where_str = ''

        if len(q):
            where_str = 'where %s' % q

        c.execute('select `%s` ' % ('`,`'.join(cols)) +
                  'from map join files on map.file_id=files.id %s' % where_str, p)
        rows = c.fetchall()
        rows = [self.scheme._format_row(dict(zip(rp, r))) for r in rows]
        if prefix is not None:
            for r in rows:
                r['filename'] = os.path.join(prefix, r['filename'])
        if filename:
            # manual filter...
            rows = [r for r in rows if r['filename'] == filename]
        return rows

    def add_entry(self, params, filename=None, create=True, commit=True,
                  replace=False):
        """Add an entry to the map table.

        Arguments:
          params: a dict of values for the Index Data columns.  In the
            case of 'range' columns, a pair of values must be
            provided.  Endpoint Data, other than the filename, should
            also be included here.
          filename: the filename to associate with matching Index Data.
          create: if False, do not create new entry in the file table
            (and fail if entry for filename does not already exist).
          commit: if False, do not commit changes to the database (for
            batch use).
          replace: if True, do not raise an error if the index data
            matches a row of the table already; instead just update
            the record.

        Notes:
          The uniqueness check in the database will only prevent (or
          replace) *identical* index entries.  Other inconsistent
          states, such as overlapping time ranges that would both
          match some single timestamp, are not caught here.

        """
        # Validate the input data.
        q, p = self.scheme.get_insertion_query(params)
        file_id = self._get_file_id(filename, create=create)
        assert(file_id is not None)
        c = self.conn.cursor()
        marks = ','.join('?' * len(p))
        c = self.conn.cursor()
        query = 'into map (%s,file_id) values (%s,?)' % (q, marks)
        if replace:
            query = 'insert or replace ' + query
        else:
            query = 'insert ' + query
        c.execute(query, p + (file_id,))
        if commit:
            self.conn.commit()

    def update_entry(self, params, filename=None, commit=True):
        """Update an existing entry.

        Arguments:
          params: Index data to change.  This must include key '_id',
            with the value corresponding to an existing row in the
            table.

        Notes:

          Only columns expressly included in params will be updated.
          The params can include 'filename', in which case a new value
          is set.

        """
        # Validate the input data.
        q, p = self.scheme.get_update_query(params)
        _id = params.get('_id')
        # Are we changing the filename?  Hope not ...
        assert(filename is None and 'filename' not in params)
        if filename is None:
            filename = params.get('filename')

        c = self.conn.cursor()
        marks = ','.join('?' * len(p))
        c = self.conn.cursor()
        query = 'update map set %s where id=?' % q
        c.execute(query, p + (_id,))# + (file_id,))
        if commit:
            self.conn.commit()

    def remove_entry(self, _id, commit=True):
        """Remove the entry identified by row id _id.

        If _id is a dict, _id['_id'] is used.  Entries returned by
        .inspect() should have _id populated in this way, and thus can
        be passed directly into this function.

        """
        if isinstance(_id, dict):
            _id = _id['_id']
        c = self.conn.cursor()
        file_id = c.execute('select file_id from map where id=?',
                             (_id, )).fetchall()
        if len(file_id) == 0:
            raise ValueError(f'Row with id={_id} does not exist!')
        file_id = file_id[0][0]
        c.execute('delete from map where id=?', (_id, ))
        if len(c.execute('select id from map where file_id=?', (file_id, )).fetchall()) == 0:
            c.execute('delete from files where id=?', (file_id, ))
        if commit:
            self.conn.commit()

    def get_entries(self, fields):
        """Return unique combinations of data fields (endpoint data) in the
        database.

        Arguments
        ---------
        fields: list of strings
            Should correspond to columns in map table made through
            ManifestScheme.add_data_field(field_name). (For 'range'
            fields, include __lo or __hi.)

        Returns
        --------
        ResultSet with keys equal to field names.

        """
        # Make sure all fields are quoted.
        fields = [(f if f[0] in ["`", "'", '"'] else f"`{f}`")
                  for f in fields]
        if not isinstance(fields, list):
            raise ValueError("fields must be a list")
        q = f"select distinct {','.join(fields)} from map"
        c = self.conn.execute(q)
        return resultset.ResultSet.from_cursor(c)
        
    def validate(self):
        """
        Checks that the database is following internal rules.
        Specifically...

        Raises SchemaError in the first case, IntervalError in the second.
        """
        return False


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(
            epilog="""For details of individual modes, pass a dummy database argument
            followed by the mode and -h, e.g.: "%(prog)s x files -h" """)

    parser.add_argument('filename', help="Path to a ManifestDb.",
                        metavar='my_db.sqlite')

    cmdsubp = parser.add_subparsers(
        dest='mode')

    # "summary"
    p = cmdsubp.add_parser(
        'summary', help=
        "Summarize database structure and number of entries.",
        usage="""

    %(prog)s

        This will print a summary of the index fields and
        endpoint fields.  (This mode is chosen by default.)
        """)

    # "entries"
    p = cmdsubp.add_parser(
        'entries', help=
        "Show all entries in the database.",
        usage="""Syntax:

    %(prog)s

        This will print every row of the metadata map table,
        including the filename, and with two header rows.
        """)


    # "files"
    p = cmdsubp.add_parser(
        'files', usage="""Syntax:

    %(prog)s
    %(prog)s --all
    %(prog)s --clean

        This will print out a list of archive files referenced
        by the database.  --all prints all the rows, even if
        there are a lot of them.  --clean is used to get a
        simple list, one file per line, for use with rsync
        or whatever.
        """,
        help="List the files referenced in the database.")

    p.add_argument('--clean', action='store_true',
                   help="Print a simple list of all files (for script digestion).")
    p.add_argument('--all', action='store_true',
                   help="Print all files, not an abbreviated list.")

    # "lookup"
    p = cmdsubp.add_parser(
        'lookup', help=
        "Query database for specific index data and display matched endpoint data.",
        usage="""Syntax:

    %(prog)s val1,val2,... [val1,val2,... ]

        Each command line argument is converted to a single query, and
        must consist of comma-delimited fields to be associated
        one-to-one with the index fields.

    Example 1: if the single index field is "obs:obs_id":

        %(prog)s obs_1840300000

    Example 2: do two queries (single index field)

        %(prog)s obs_1840300000 obs_185040012

    Example 3: single query, two index fields (perhaps timestamp and wafer):

        %(prog)s 1840300000,wafer29

        """)
    p.add_argument('index', nargs='*', help=
                   "Index information.  Comma-delimit your data, for example: "
                   "1456453245,wafer5")

    # "reroot"
    p = cmdsubp.add_parser(
        'reroot', help=
        "Batch change filenames (by prefix) in the database.",
        usage="""Syntax:

    %(prog)s old_prefix new_prefix [output options]

Examples:

    %(prog)s /path/on/system1 /path/on/system2 -o my_new_manifest.sqlite
    %(prog)s /path/on/system1 /new_path/on/system1 --overwrite
    %(prog)s ./result1/obs_12345.h5 ./result2/obs_12345.h5 --overwrite

        These operations will create a new ManifestDb, will all the
        entries from my_db.sqlite, but with the filenames
        (potentially) altered.  Any filename that starts with the
        first argument will be changed, in the output, to instead
        start with the second argument.  When you do this you must
        either say where to write the output (-o) or give the program
        permission to overwrite your input database file.  Note that
        the first argument need not match all entries in the database;
        you can use it to pick out a subset (even a single entry).
    """)
    p.add_argument('old_prefix', help=
                   "Prefix to match in current database.")
    p.add_argument('new_prefix', help=
                   "Prefix to replace it with.")
    p.add_argument('--overwrite', action='store_true', help=
                   "Store modified database in the same file.")
    p.add_argument('--output-db', '-o', help=
                   "Store modified database in this file.")
    p.add_argument('--dry-run', action='store_true', help=
                   "Run the conversion steps but do not write the results anywhere.")

    return parser

def main(args=None):
    """Entry point for the so-metadata tool."""
    if args is None:
        args = sys.argv[1:]
    if not isinstance(args, argparse.Namespace):
        parser = get_parser()
        args = parser.parse_args(args)

    if args.mode is None:
        args.mode = 'summary'

    db = ManifestDb.from_file(args.filename, force_new_db=False)

    if args.mode == 'summary':
        header = f'Summary for {args.filename}'
        print(header)
        print('-' * len(header))
        print()

        schema = db.scheme.as_resultset()
        row_count = db.conn.execute('select count(id) from map').fetchone()[0]
        print(f'Total number of index entries:  {row_count:>7}')
        print()

        print('Index fields:')
        fmt = '   {field:20}  {data_type:20} {col_type:>10}'
        hdr = fmt.format(field="Field", data_type="Type", col_type="Match")
        print(hdr)
        print('   ' + '-' * (len(hdr) - 3))
        for row in schema:
            if row['purpose'] == 'in':
                print(fmt.format(**row))
        print()

        print('Endpoint fields:')
        fmt = '   {field:20}  {data_type:20} {count:>10}'
        hdr = fmt.format(field="Field", data_type="Type", count="Entries")
        print(hdr)
        print('   ' + '-' * (len(hdr) - 3))
        for row in schema:
            if row['purpose'] == 'out':
                count = len(db.conn.execute("select distinct `%s` from map" % row['field']).fetchall())
                print(fmt.format(count=count, **row))
        file_count = db.conn.execute('select count(id) from files').fetchone()[0]
        print(fmt.format(
            field='filename', data_type='filename', count=file_count))

        print()

    elif args.mode == 'entries':
        # Print the table of entries.
        schema = db.scheme.as_resultset()
        keys = []
        for purp in ['in', 'out']:
            for s in schema:
                if s['purpose'] != purp:
                    continue
                if s['col_type'] == 'range':
                    keys.extend([s['field'] + '__lo',
                                 s['field'] + '__hi'])
                else:
                    keys.append(s['field'])
        keys.append('filename')
        print(keys)
        print(['-'] * len(keys))
        for row in db.conn.execute('select map.*, files.name as filename from '
                                   'map join files where map.file_id=files.id'):
            print([row[k] for k in keys])

    elif args.mode == 'files':
        # Get all files.
        rows = db.conn.execute(
            'select files.id, files.name as filename, count(map.id) '
            'from map join files on '
            'map.file_id==files.id group by filename').fetchall()

        if args.clean:
            for _id, filename, count in rows:
                print(filename)
        else:
            fmt = '  {_id:>7} {count:>7} {filename}'
            hdr = fmt.format(_id="file_id", count="Count", filename="Filename")
            print(hdr)
            print('-' * (len(hdr) + 20))
            n = len(rows)
            super_count = sum([r[2] for r in rows])
            if n > 20 and not args.all:
                rows = rows[:10]
            for _id, filename, count in rows:
                print(fmt.format(_id=_id, filename=filename, count=count))
            if len(rows) < n:
                other_count = super_count - sum([r[1] for r in rows])
                print(fmt.format(count=other_count, filename='...'))
                print('(Pass --all to show all results.)')
            print()

    elif args.mode == 'lookup':
        schema = db.scheme.as_resultset()

        index_fields = [r['field'] for r in schema if r['purpose'] == 'in']
        endpoint_fields = [r['field'] for r in schema if r['purpose'] == 'out']

        results = []
        for index in args.index:
            vals = index.split(',')
            if len(vals) != len(index_fields):
                print(f'Index data "{index}" decodes to "{len(vals)}" fields, ')
                print(f'but we expected "{len(index_fields)}".')

            query = {r: a for r, a in zip(index_fields, vals)}
            matches = db.match(query, multi=True)

            results.append({'query': query,
                            'matches': []})
            endpoint_fields.append('filename')
            for i, m in enumerate(matches):
                results[-1]['matches'].append({})
                for k in endpoint_fields:
                    results[-1]['matches'][-1][k] = matches[i][k]

        print(json.dumps(results, indent=4))

    elif args.mode == 'reroot':
        # Reconnect with write?
        if args.overwrite:
            if args.output_db:
                parser.error("Specify only one of --overwrite or --output-db.")
            db = ManifestDb.from_file(args.filename, force_new_db=True)
            args.output_db = args.filename
        else:
            if args.output_db is None:
                parser.error("Specify an output database name with --output-db, "
                             "or pass --overwrite to clobber.")
            db = ManifestDb.from_file(args.filename, force_new_db=True)

        # Get all files matching this prefix ...
        c = db.conn.execute("select id, name from files "
                            "where name like '%s%%'" % (args.old_prefix))
        rows = c.fetchall()
        print('Found %i records matching prefix ...'
               % len(rows))

        print('Converting to new prefix ...')
        n_examples = 1

        if not args.dry_run:
            c = db.conn.cursor()

        for (_id, name) in rows:
            new_name = args.new_prefix + name[len(args.old_prefix):]
            if n_examples > 0:
                print(f'  Example: converting filename\n'
                      f'      "{name}"\n'
                      f'    to\n'
                      f'      "{new_name}"')
                n_examples -= 1
            if not args.dry_run:
                c.execute('update files set name=? where id=?', (new_name, _id))

        print('Saving to %s' % args.output_db)
        if not args.dry_run:
            db.conn.commit()
            c.execute('vacuum')
            db.to_file(args.output_db)

    else:
        print(f'Sorry, {args.mode} not implemented.')



# Copyright (c) 2020-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Tools for saving Level-3 Book format data.

"""

import os
import sys
from datetime import datetime, timezone

import yaml
import numpy as np
from astropy import units as u
from astropy.table import Table, QTable
import h5py

# Import so3g before any other packages that import spt3g
import so3g
from spt3g import core as c3g

import toast
import toast.spt3g as t3g
import toast.qarray as qa
from toast.observation import default_values as defaults

from ...sim_hardware import telescope_tube_wafer
from ...__init__ import __version__ as sotodlib_version


def obs_wafer_list(obs_list):
    """Get the list of wafers covered by a set of observations.

    Args:
        obs_list (list):  The list of observations.

    Returns:
        (tuple):  The (tube name, list of wafers).

    """
    tube_check = None
    wafers = list()

    for obs in obs_list:
        focalplane = obs.telescope.focalplane
        props = focalplane.detector_data

        obs_tube_check = set(props[:]["tube_slot"])
        if len(obs_tube_check) != 1:
            msg = f"Observation has multiple tube_slots ({obs_tube_check})"
            raise RuntimeError(msg)
        tube_name = obs_tube_check.pop()

        if tube_check is None:
            tube_check = tube_name
        else:
            if tube_name != tube_check:
                msg = f"Observations in session have multiple tubes"
                raise RuntimeError(msg)

        wafer_set = set(props[:]["wafer_slot"])
        for wf in wafer_set:
            if wf not in wafers:
                wafers.append(wf)
    return tube_check, wafers


def book_name_from_obs(obs_list):
    """Return the name of the book, given a list of observations.

    This uses focalplane information to build the name of the book directory.  The
    observations should all be from the same observing session.

    Args:
        obs_list (list):  The list of observations.

    Returns:
        (str):  The book name.

    """
    log = toast.utils.Logger.get()
    timestamp = int(obs_list[0].session.start.timestamp())

    ttw = telescope_tube_wafer()
    tele_name = obs_list[0].telescope.focalplane.detector_data.meta["telescope"]
    tube_name, wafers = obs_wafer_list(obs_list)

    if tele_name == "LAT":
        tele_tube = f"lat{tube_name}"
    else:
        tele_tube = tele_name

    wafer_bitstr = ""
    for wf in ttw[tele_name][tube_name]:
        if wf in wafers:
            wafer_bitstr = f"{wafer_bitstr}1"
        else:
            wafer_bitstr = f"{wafer_bitstr}0"
    return f"obs_{timestamp:10d}_{tele_tube}_{wafer_bitstr}"


def format_book_time(dt):
    """Format a datetime in the desired way for books.

    Args:
        dt (datetime):  A timezone-aware datetime object.

    Returns:
        (str):  The formatted time string.

    """
    return datetime.strftime(dt, "%Y-%m-%dT%H:%M:%S%z")


def write_frame_files(comm, frames, froot, gzip=False):
    fname = f"{froot}_{comm.group_rank:03d}.g3"
    tempname = f"{froot}_{comm.group_rank:03d}_temp.g3"
    if gzip:
        fname += ".gz"
        tempname += ".gz"
    writer = c3g.G3Writer(tempname)
    for frm in frames:
        # if frm.type == c3g.G3FrameType.Scan and "signal" in frm:
        #     print(f"PROCESS {frm['signal'].data}")
        writer.Process(frm)
    del writer
    if comm.comm_group is not None:
        # Wait for all writers to finish
        comm.comm_group.barrier()
    # Move temp file into place
    os.rename(tempname, fname)


def export_obs_meta(obs):
    """Extract Observation metadata.

    Return general metadata as an Observation frame and a corresponding
    dictionary suitable for writing to the book metadata.

    Args:
        obs (Observation):  One observation in the session.

    Returns:
        (tuple):  The (frame, dictionary).

    """
    log = toast.utils.Logger.get()

    # Construct observation frame and dictionary
    ob = c3g.G3Frame(c3g.G3FrameType.Observation)
    obmeta = dict()

    session = obs.session
    log.verbose(f"Create observation frame for session {session.name}")

    ob["telescope_name"] = c3g.G3String(
        obs.telescope.focalplane.detector_data.meta["telescope"]
    )
    obmeta["telescope_name"] = str(
        obs.telescope.focalplane.detector_data.meta["telescope"]
    )
    ob["telescope_uid"] = c3g.G3Int(obs.telescope.uid)
    obmeta["telescope_uid"] = int(obs.telescope.uid)

    ob["observing_session"] = c3g.G3String(session.name)
    obmeta["observing_session"] = str(session.name)

    ob["observing_session_uid"] = c3g.G3Int(session.uid)
    obmeta["observing_session_uid"] = int(session.uid)

    ob["observing_session_start"] = t3g.to_g3_time(session.start.timestamp())
    obmeta["observing_session_start"] = float(session.start.timestamp())

    ob["observing_session_end"] = t3g.to_g3_time(session.end.timestamp())
    obmeta["observing_session_end"] = float(session.end.timestamp())

    site = obs.telescope.site
    siteclass = toast.utils.object_fullname(site.__class__)
    ob["site_name"] = c3g.G3String(site.name)
    ob["site_class"] = c3g.G3String(siteclass)
    ob["site_uid"] = c3g.G3Int(site.uid)
    obmeta["site_name"] = str(site.name)
    obmeta["site_class"] = str(siteclass)
    obmeta["site_uid"] = int(site.uid)

    if isinstance(site, toast.instrument.GroundSite):
        ob["site_lat_deg"] = c3g.G3Double(site.earthloc.lat.to_value(u.degree))
        ob["site_lon_deg"] = c3g.G3Double(site.earthloc.lon.to_value(u.degree))
        ob["site_alt_m"] = c3g.G3Double(site.earthloc.height.to_value(u.meter))
        obmeta["site_lat_deg"] = float(site.earthloc.lat.to_value(u.degree))
        obmeta["site_lon_deg"] = float(site.earthloc.lon.to_value(u.degree))
        obmeta["site_alt_m"] = float(site.earthloc.height.to_value(u.meter))

        if site.weather is not None:
            if hasattr(site.weather, "name"):
                # This is a simulated weather object, dump it.
                ob["site_weather_name"] = c3g.G3String(site.weather.name)
                obmeta["site_weather_name"] = str(site.weather.name)
                ob["site_weather_realization"] = c3g.G3Int(site.weather.realization)
                obmeta["site_weather_realization"] = int(site.weather.realization)
                if site.weather.max_pwv is None:
                    ob["site_weather_max_pwv"] = c3g.G3String("NONE")
                    obmeta["site_weather_max_pwv"] = "NONE"
                else:
                    ob["site_weather_max_pwv"] = c3g.G3Double(site.weather.max_pwv)
                    obmeta["site_weather_max_pwv"] = float(site.weather.max_pwv)
                ob["site_weather_time"] = t3g.to_g3_time(site.weather.time.timestamp())
                obmeta["site_weather_time"] = float(site.weather.time.timestamp())
                ob["site_weather_uid"] = c3g.G3Int(site.weather.site_uid)
                obmeta["site_weather_uid"] = int(site.weather.site_uid)
                ob["site_weather_use_median"] = c3g.G3Bool(site.weather.median_weather)
                obmeta["site_weather_use_median"] = bool(site.weather.median_weather)
    return ob, obmeta


def export_obs_ancil(
    obs,
    book_id,
    times,
    shared_flags,
    boresight_azel,
    boresight_radec=None,
    corotator_angle=None,
    boresight_angle=None,
    hwp_angle=None,
    frame_intervals=None,
):
    """Extract Observation ancillary data.

    This extracts the shared data fields.  The input observation is one within the
    observing session, since all such observations are guaranteed to have the same
    telescope pointing and flags.

    Args:
        obs (Observation):  One observation in the session.
        book_id (str):  The book ID.
        times (str):  The field containing timestamps.
        shared_flags (str):  The field containing shared flags.
        boresight_azel (str):  The boresight Az / El quaternions.
        boresight_radec (str):  The boresight RA / DEC quaternions.
        corotator_angle (str):  The field containing the corotator angle.
        boresight_angle (str):  The field containing the boresight rotation angle.
        hwp_angle (str):  The field containing the HWP rotation angle.
        frame_intervals (str):  The name of the intervals to use for frame boundaries.
            If not specified, the observation sample sets are used.

    Returns:
        (list):  The frames.

    """
    log = toast.utils.Logger.get()
    log.verbose(f"Create ancillary frames for {book_id}")

    output = list()
    for ifrm, frmint in enumerate(obs.intervals[frame_intervals]):
        slc = slice(frmint.first, frmint.last, 1)
        # Construct the Scan frame
        frame = c3g.G3Frame(c3g.G3FrameType.Scan)
        frame["book_id"] = c3g.G3String(book_id)
        frame["sample_range"] = c3g.G3VectorInt(
            np.array(
                [frmint.first, frmint.last],
                dtype=np.int64,
            )
        )

        ancil = c3g.G3TimesampleMap()
        ancil.times = t3g.to_g3_time(obs.shared[times].data[slc])

        theta, phi, pa = qa.to_iso_angles(obs.shared[boresight_azel].data[slc, :])
        ancil["az_enc"] = c3g.G3VectorDouble(-phi)
        ancil["el_enc"] = c3g.G3VectorDouble((np.pi / 2) - theta)
        # ancil["flags"] = c3g.G3VectorUnsignedChar(obs.shared[shared_flags].data[slc])
        ancil["flags"] = c3g.G3VectorInt(
            obs.shared[shared_flags].data[slc].astype(np.int32)
        )
        frame["qboresight_azel"] = t3g.to_g3_quats(
            obs.shared[boresight_azel].data[slc, :]
        )
        if boresight_radec is not None:
            frame["qboresight_radec"] = t3g.to_g3_quats(
                obs.shared[boresight_radec].data[slc, :]
            )

        if hwp_angle is not None and hwp_angle in obs.shared:
            ancil["hwp_enc"] = c3g.G3VectorDouble(obs.shared[hwp_angle].data[slc])

        if corotator_angle is not None and corotator_angle in obs.shared:
            ancil["corotation_enc"] = c3g.G3VectorDouble(
                obs.shared[corotator_angle].data[slc]
            )

        if boresight_angle is not None and boresight_angle in obs.shared:
            ancil["boresight_enc"] = c3g.G3VectorDouble(
                obs.shared[boresight_angle].data[slc]
            )

        frame["ancil"] = ancil
        output.append(frame)

    return output


def export_obs_data(
    wafer_obs,
    full_fp,
    times,
    det_data,
    det_flags,
    frame_intervals=None,
):
    """Extract Observation detector data.

    This returns just the frame objects that are detector related.  The ancil
    frame data is merged in at a higher level.  The wafer_obs list is a subset
    of observations in the session which contain detectors at multiple frequencies
    within the same wafer.  Any readout_ids that are not included in the observations
    have their data set to zero and their flags set to invalid.

    Args:
        wafer_obs (list):  The observations spanning a single wafer.
        full_fp (SOFocalplane):  A focalplane instance containing all detectors in the
            session.
        times (str):  The field containing timestamps.
        det_data (str):  The detector data name to export.
        det_flags (str):  The detector flags name to export.
        frame_intervals (str):  The name of the intervals to use for frame boundaries.
            If not specified, the observation sample sets are used.

    Returns:
        (list):  The frames.

    """
    log = toast.utils.Logger.get()

    obs_names = [x.name for x in wafer_obs]
    log.verbose(f"Create data frames for {obs_names}")

    wafer_check = set()
    for ob in wafer_obs:
        wob = set(ob.telescope.focalplane.detector_data[:]["wafer_slot"])
        wafer_check.update(wob)
    if len(wafer_check) != 1:
        msg = f"Observation list {obs_names} contains multiple wafers ({wafer_check})"
        log.error(msg)
        raise RuntimeError(msg)

    the_wafer = wafer_check.pop()

    # Extract just the detectors in this wafer
    wafer_rows = full_fp.detector_data["wafer_slot"] == the_wafer
    wafer_data = full_fp.detector_data[wafer_rows]

    det_names = wafer_data["readout_id"]

    # For each observation, compute the indexing from the detectors into the full-wafer
    # data

    wafer_indices = list()
    full_name_to_row = {y: x for x, y in enumerate(wafer_data["name"])}
    for ob in wafer_obs:
        wafer_indices.append(
            np.array([full_name_to_row[x] for x in ob.all_detectors], dtype=np.int32)
        )

    # We are going to walk through the observations in lock step and for each frame
    # we will extract the relevant detectors.  Since all of these observations should
    # have identical frame boundaries, we get those from the first observation.

    output = list()

    for ifrm, frmint in enumerate(wafer_obs[0].intervals[frame_intervals]):
        slc = slice(frmint.first, frmint.last, 1)

        # Get the timestamps from the first observation
        g3times = t3g.to_g3_time(wafer_obs[0].shared[times].data[slc])

        # Construct the Scan frame
        frame = c3g.G3Frame(c3g.G3FrameType.Scan)
        frame["stream_id"] = c3g.G3String(wafer_data["wafer_slot"][0])

        # Create timestream containers for detector data and flags.
        # FIXME:  we could eventually enable compression here, but first
        # need to decide how we want to reduce our precision and we should
        # modify unit tests to perform the same operations in memory on
        # the input data.

        dts = so3g.G3SuperTimestream()
        dts.names = det_names
        dts.times = g3times
        dts.quanta = np.ones(len(det_names))
        dts.options(enable=0)
        dts_temp = np.zeros(
            (len(det_names), frmint.last - frmint.first),
            dtype=wafer_obs[0].detdata[det_data].dtype,
        )

        fts = so3g.G3SuperTimestream()
        fts.names = det_names
        fts.times = g3times
        fts.options(enable=0)
        fts_temp = np.zeros(
            (len(det_names), frmint.last - frmint.first), dtype=np.int32
        )

        # Timestream units
        utype, uscale = t3g.to_g3_unit(ob.detdata[det_data].units)
        frame["signal_units"] = utype

        # Fill the data
        for iob, ob in enumerate(wafer_obs):
            for idet, det in enumerate(ob.all_detectors):
                dts_temp[wafer_indices[iob][idet], :] = ob.detdata[det_data][
                    idet, slc
                ] * uscale
                fts_temp[wafer_indices[iob][idet], :] = ob.detdata[det_flags][
                    idet, slc
                ]

        dts.data = dts_temp
        # dts.options(enable=1)
        frame["signal"] = dts
        del dts

        fts.data = fts_temp
        # fts.options(enable=1)
        frame["flags"] = fts
        del fts

        output.append(frame)
    return output


@toast.timing.function_timer
def write_book(
    obs_list,
    full_fp,
    book_dir,
    focalplane_dir,
    noise_dir,
    times,
    shared_flags,
    det_data,
    det_flags,
    boresight_azel,
    boresight_radec,
    corotator_angle,
    boresight_angle,
    hwp_angle,
    gzip=False,
    frame_intervals=None,
    redist_sampsets=False,
):
    """Write a set of observations as a book.

    An observation book is created inside the parent `book_dir`.  The observations
    passed in should be from the same observing "session".  For Simons Observatory,
    a TOAST session is mapped to a telescope tube of data for some time span.

    When building the book, observations from the same wafer at different frequencies
    are combined to write the detector data files for that wafer.  If data for one
    frequency is missing, the timestreams for those detectors is set to zero.

    To-Do:  We could add more options to dump out multiple flavors of detector
    data rather than just the defaults.

    Args:
        obs_list (list):  The list of observations.
        full_fp (SOFocalplane):  A focalplane instance containing all detectors in the
            session.
        book_dir (str):  The parent directory of the output book.
        focalplane_dir (str):  Path to directory for dumping per-wafer focalplane info.
        noise_dir (str):  Path to directory for dumping per-wafer noise models.
        times (str):  The field containing timestamps.
        shared_flags (str):  The field containing shared flags.
        det_data (str):  The detector data name to export.
        det_flags (str):  The detector flags name to export.
        boresight_azel (str):  The boresight Az / El quaternions.
        boresight_radec (str):  The boresight RA / DEC quaternions.
        corotator_angle (str):  The field containing the corotator angle.
        boresight_angle (str):  The field containing the boresight rotation angle.
        hwp_angle (str):  The field containing the HWP rotation angle.
        gzip (bool):  If True, gzip the frame files.
        frame_intervals (str):  The intervals to use for frame boundaries.
        redist_sampsets (list):  When redistributing, use these new sample sets.

    Returns:
        None.

    """
    log = toast.utils.Logger.get()
    if len(obs_list) == 0:
        # Nothing to do!
        log.warning("List of observations to write is empty, skipping")
        return

    # All the observations are on the same group
    comm = obs_list[0].comm

    # Get the current time as the "book finalization" time.
    now = None
    if comm.group_rank == 0:
        now = datetime.now(tz=timezone.utc)
    if comm.comm_group is not None:
        now = comm.comm_group.bcast(now)

    # Get the start and end time of the session
    timestamp_start = None
    timestamp_end = None
    session = obs_list[0].session
    if session.start is None or session.end is None:
        # We have to get the time range from the observations
        if comm.group_rank == 0:
            timestamp_start = obs_list[0].shared[times][0]
        if comm.group_rank == comm.group_size - 1:
            timestamp_end = obs_list[0].shared[times][-1]
    else:
        # Use the session times
        timestamp_start = session.start.timestamp()
        timestamp_end = session.end.timestamp()
    if comm.comm_group is not None:
        timestamp_start = comm.comm_group.bcast(timestamp_start, root=0)
        timestamp_end = comm.comm_group.bcast(timestamp_end, root=(comm.group_size - 1))

    log.verbose_rank(
        f"Writing book for session {session.name} with {len(obs_list)} observations",
        comm=comm.comm_group,
    )

    # Get the observation frame from the first observation of the session
    obs_frame, obs_props = export_obs_meta(obs_list[0])

    # Redistribute observations so that every process has a time slice
    for obs in obs_list:
        obs.redistribute(
            1,
            times=times,
            override_sample_sets=redist_sampsets,
        )

    # Create book directory
    book_name = book_name_from_obs(obs_list)
    book_path = os.path.join(book_dir, book_name)
    if comm.group_rank == 0:
        if not os.path.isdir(book_path):
            os.makedirs(book_path)
    if comm.comm_group is not None:
        comm.comm_group.barrier()

    # Get the ancillary frames
    ancil_frames = export_obs_ancil(
        obs_list[0],
        book_name,
        times,
        shared_flags,
        boresight_azel,
        boresight_radec,
        corotator_angle=corotator_angle,
        boresight_angle=boresight_angle,
        hwp_angle=hwp_angle,
        frame_intervals=frame_intervals,
    )

    # Collect the total sample ranges of all frames in each
    # file, for writing into the index later.
    local_sample_ranges = [
        int(ancil_frames[0]["sample_range"][0]),
        int(ancil_frames[-1]["sample_range"][1]),
    ]
    if comm.comm_group is None:
        file_sample_ranges = [local_sample_ranges]
    else:
        file_sample_ranges = comm.comm_group.gather(local_sample_ranges, root=0)

    # Write ancillary frames
    froot = os.path.join(book_path, "A_ancil")
    write_frames = [obs_frame]
    write_frames.extend(ancil_frames)
    write_frame_files(comm, write_frames, froot, gzip=gzip)

    # Split up observations into lists per wafer
    all_wafers = dict()
    for ob in obs_list:
        wf = ob.telescope.focalplane.detector_data["wafer_slot"][0]
        if wf not in all_wafers:
            all_wafers[wf] = list()
        all_wafers[wf].append(ob)

    # Export per-wafer frames
    for wf, wafer_obs in all_wafers.items():
        det_frames = export_obs_data(
            wafer_obs,
            full_fp,
            times,
            det_data,
            det_flags,
            frame_intervals=frame_intervals,
        )
        # Add the ancillary data into the detector frames
        for aframe, dframe in zip(ancil_frames, det_frames):
            dframe["book_id"] = aframe["book_id"]
            dframe["sample_range"] = aframe["sample_range"]
            dframe["ancil"] = aframe["ancil"]
            dframe["qboresight_azel"] = aframe["qboresight_azel"]
            if "qboresight_radec" in aframe:
                dframe["qboresight_radec"] = aframe["qboresight_radec"]
        # Write
        froot = os.path.join(book_path, f"D_{wf}")
        write_frames = [obs_frame]
        write_frames.extend(det_frames)
        write_frame_files(comm, write_frames, froot, gzip=gzip)

    del write_frames
    del det_frames
    del ancil_frames

    # Write out remaining metadata

    if comm.group_rank == 0:
        if not os.path.isdir(focalplane_dir):
            os.makedirs(focalplane_dir)
        if noise_dir is not None:
            if not os.path.isdir(noise_dir):
                os.makedirs(noise_dir)
    if comm.comm_group is not None:
        comm.comm_group.barrier()

    for wf, wafer_obs in all_wafers.items():
        for ob in wafer_obs:
            fp = ob.telescope.focalplane
            band = fp.detector_data["band"][0]
            fp_file = os.path.join(focalplane_dir, f"{book_name}_{wf}_{band}.h5")
            with toast.io.H5File(fp_file, "w", comm=comm.comm_group) as f:
                fp.save_hdf5(f.handle, comm=comm.comm_group)
            # Go through the observation and find any noise models
            for k in list(sorted(ob.keys())):
                v = ob[k]
                if isinstance(v, toast.noise.Noise):
                    noise_file = os.path.join(
                        noise_dir, f"{book_name}_{wf}_{band}_{k}.h5"
                    )
                    with toast.io.H5File(noise_file, "w", comm=comm.comm_group) as f:
                        v.save_hdf5(f.handle, ob)

    # Write out observation metadata as a yaml file, in addition to the observation
    # frame.  NOTE:  we copy what is done in the BookBinder class.
    if comm.group_rank == 0:
        obs_meta = dict()
        obs_meta["obs_id"] = book_name
        obs_meta["type"] = "obs"
        obs_meta["book_id"] = book_name
        obs_meta["observatory"] = "Simons Observatory"
        obs_meta["telescope"] = obs_props["telescope_name"]
        obs_meta["stream_ids"] = [str(x) for x in all_wafers.keys()]
        obs_meta["detsets"] = obs_meta["stream_ids"]
        obs_meta["start_time"] = timestamp_start
        obs_meta["stop_time"] = timestamp_end
        obs_meta["timestamp_start"] = format_book_time(
            datetime.fromtimestamp(timestamp_start).astimezone(timezone.utc)
        )
        obs_meta["timestamp_end"] = format_book_time(
            datetime.fromtimestamp(timestamp_end).astimezone(timezone.utc)
        )
        obs_meta["sample_ranges"] = file_sample_ranges
        obs_meta["n_samples"] = file_sample_ranges[-1][1]
        obs_meta["tags"] = []
        obs_meta_file = os.path.join(book_path, f"M_index.yaml")
        try:
            obs_meta_content = yaml.dump(obs_meta, default_flow_style=False)
            with open(obs_meta_file, "w") as f:
                f.write(obs_meta_content)
        except yaml.YAMLError:
            log.error("Cannot write M_index.yaml")
            raise

        # Write out toast properties
        toast_meta_file = os.path.join(book_path, f"M_toast.yaml")
        try:
            toast_meta_content = yaml.dump(obs_props, default_flow_style=False)
            with open(toast_meta_file, "w") as f:
                f.write(toast_meta_content)
        except yaml.YAMLError:
            log.error("Cannot write M_toast.yaml")
            raise

        # Write out book metadata last- so we can use the existence of this file as
        # an indicator of success
        meta = dict()
        book_meta = dict()
        book_meta["type"] = "obs"
        book_meta["book_id"] = book_name
        book_meta["schema_version"] = 0
        book_meta["finalized_at"] = format_book_time(now)
        meta["book"] = book_meta
        binder_meta = dict()
        binder_meta["codebase"] = "sotodlib.toast"
        binder_meta[
            "version"
        ] = f"sotodlib={sotodlib_version},toast={toast.__version__}"
        binder_meta["context"] = "testing"
        meta["bookbinder"] = binder_meta
        meta_file = os.path.join(book_path, f"M_book.yaml")
        try:
            meta_content = yaml.dump(meta, default_flow_style=False)
            with open(meta_file, "w") as f:
                f.write(meta_content)
        except yaml.YAMLError:
            log.error("Cannot write M_book.yaml")
            raise

    if comm.comm_group is not None:
        comm.comm_group.barrier()



import os
import yaml

from ..core.util import tag_substr

def get_tags(env_file=None, env_var="DATAPKG_ENV"):
    if env_file is None:
        env_file = os.environ.get(env_var)
    if env_file is None:
        tags = {}
    else:
        tags = yaml.safe_load( open(env_file, "r") )
    return tags

def load_configs(cfg_file, env_file=None, env_var="DATAPKG_ENV"):
    """ load a yaml config file and replace string paths tags
    
    Example config file with a tag::

        my_path : /complete/path
        other_path: {tag}/path
    
    matches with an environment file which would contain::

        tag: /other/relevant/thing

    so that the returned config dictionary contains::

        'other_path' : '/other/relevant/thing/path'
    
    Arguments
    ----------
    cfg_file: str
        path to a .yaml configuration file
    env_file: str (optional)
        path to a .yaml file where each entry is a possible tag.
    env_var: str (optional)
        

    Returns
    --------
    configs: dict
        loaded configuration dictionary where all strings with tags have been
        replaced with the values loaded from the environment file
    """
    tags = get_tags(env_file=env_file, env_var=env_var)

    configs = yaml.safe_load( open(cfg_file, "r"))
    try:
        configs = tag_substr(configs, tags)
    except KeyError as e:
        raise ValueError(
            f"Config file {cfg_file} expects tag {e.args[0]} but this tag is "
            f"not found in environment file '{env_file}'"
        )
    return configs


def get_imprinter_config( platform, env_file=None, env_var="DATAPKG_ENV"):
    tags = get_tags(env_file=env_file, env_var=env_var)

    if 'configs' not in tags:
        raise ValueError(f"configs not found in tags {tags}")

    return os.path.join( tags['configs'], platform, 'imprinter.yaml')

def walk_files(path, include_suprsync=False):
    """get a list of the files in a timecode folder, optional flag to ignore
    suprsync files
    
    Arguments
    ----------
    path: path to a level 2 timecode folder, either smurf or timestreams
    include_suprsync: optional, bool
        if true, includes the suprsync files in the returned list

    Returns
    --------
    files (list): list of the absolute paths to all files in a timecode folder
    """
    if not os.path.exists(path):
        return []
    flist = []
    for root, _, files in os.walk(path):
        if not include_suprsync and 'suprsync' in root:
            continue
        for f in files:
            flist.append( os.path.join(path, root, f))
    return flist

def just_suprsync(path):
    """check if timecode folder only has suprsync folder in it
    """
    flist = os.listdir( path )
    if len(flist) == 1 and flist[0] == "suprsync":
        return True
    return False



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Data Filtering

This module contains code for filtering data in G3Frames

"""
import numpy as np
import scipy.signal as signal

from spt3g import core
from .core.g3_core import DataG3Module

class Filter(DataG3Module):
    """
    G3Module that takes the G3Timestream map and applies generic filter

    Attributes:
        input (str): the key to a G3Timestream map of the source
        output (str): key of G3Timestream map of output data.
            if None, input will be overwritten with output
        filter_function (function): function that takes frequency in Hz and
            returns a frequency filter

    TODO:
        Get rid of numpy fft functions and get faster / better parallizable
        options.
    """

    def __init__(self, input='signal', output='signal_filtered', filter_function=None):
        if filter_function is None:
            raise ValueError('Missing Filter Definition')
        self.filter_function = filter_function
        super().__init__(input, output)

    def process(self, data, det_name):
        """
        Args:
            data (G3Timestream): data for a single detector
            det_name (str): the detector name in the focal plane
                in case it's needed for accessing calibration info
        Returns:
            np.array filtered by filter_function
        """
        freqs = np.fft.rfftfreq(data.n_samples, core.G3Units.Hz/data.sample_rate)
        return np.fft.irfft( np.fft.rfft(data)*self.filter_function(freqs) )

class LowPassButterworth(Filter):
    """
    G3Module for a LowPassButterworth filter

    Attributes:
            order (int): order of butterworth
            fc (float): cutoff frequency in Hertz
            gain (float): filter gain
    """
    def __init__(self, input='signal', output='signal_filtered',
                     order=2, fc=1, gain=1):
        self.order=2
        self.fc=fc
        self.gain=gain
        super().__init__(input, output, self.filter_function)

    def filter_function(self, freqs):
        b, a = signal.butter(self.order, 2*np.pi*self.fc, 'lowpass', analog=True)
        return self.gain*np.abs(signal.freqs(b, a, 2*np.pi*freqs)[1])



from argparse import ArgumentParser
import logging
import numpy as np
import os
import sys
import yaml

import sotodlib
from sotodlib import core, coords, site_pipeline
from . import util

logger = logging.getLogger(__name__)

def get_parser(parser=None):
    if parser is None:
        parser = ArgumentParser()
    parser.add_argument('obs_id',help=
                        "Observation for which to generate flags.")
    parser.add_argument('-c', '--config-file', help=
                        "Configuration file.")
    parser.add_argument('-v', '--verbose', action='count',
                        default=0, help="Pass multiple times to increase.")
    return parser

def get_config(config_file, obs_id=None, verbose=None):
    cfg = yaml.safe_load(open(config_file, 'r'))
    for k in ['obs_id', 'verbose']:
        cfg[k] = getattr(args, k)
    return cfg

def main(config_file=None, verbose=0, obs_id=None):
    config = get_config(config_file)

    if verbose >= 1:
        logger.setLevel('INFO')
    if verbose >= 2:
        sotodlib.logger.setLevel('INFO')
    if verbose >= 3:
        sotodlib.logger.setLevel('DEBUG')

    ctx = core.Context(config['context_file'])

    # This will process data by detset, but will insist that this maps
    # one-to-one with some detdb field.
    group_by = config['subobs'].get('group_by', 'dets:detset')
    if group_by.startswith('dets:'):
        group_by = group_by.split(':',1)[1]
    else:
        raise ValueError("Can't group by '{group_by}'")

    if os.path.exists(config['archive']['index']):
        logger.info(f'Mapping {config["archive"]["index"]} for the archive index.')
        db = core.metadata.ManifestDb(config['archive']['index'])
    else:
        logger.info(f'Creating {config["archive"]["index"]} for the archive index.')
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dets:' + group_by)
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(config['archive']['index'], scheme=scheme)

    # We will loop over detsets here, but insist on a simple
    # correspondence to the output index values.
    index_map = {}

    detsets = ctx.obsfiledb.get_detsets(obs_id)
    for detset in detsets:
        logger.info(f'Loading {config["obs_id"]}:detset={detset}')
        # Load pointing and dets axis; we don't need signal though.
        tod = ctx.get_obs(obs_id, detsets=[detset],
                          no_signal=True)

        # Figure out the indexing info for this.
        this_index = tod.det_info[group_by]
        assert([x == this_index[0] for x in this_index])
        this_index = this_index[0]
        if this_index in index_map:
            raise RuntimeError(f"Some detset already mapped to {this_index}: "
                               f"{index_map[this_index]}")
        index_map[this_index] = detset
        logger.info(f'detset "{detset}" corresponds to {group_by}="{this_index}"')

        # Load / compute mask parameters.
        mask_params = config['mask_params']['default']
        mask_res = util.parse_quantity(
            util.lookup_conditional(
                config['mask_params'], 'res', default=[1, 'arcmin']),
            'deg').value

        sources = coords.planets.get_nearby_sources(tod)
        flags = None
        for source_name, eph_object in sources:
            logger.info(f'Flagging for {source_name} ...')
            _flags = coords.planets.compute_source_flags(
                tod, center_on=source_name, res=mask_res*coords.DEG,
                mask=mask_params)
            weight = np.mean(_flags.get_stats()['samples']) / _flags.shape[1]
            logger.info(f' ... weight for {source_name} was {weight*100:.2}%.')
            if flags is None:
                flags = _flags
            else:
                flags += _flags

        if flags is None:
            import so3g  # This is here to dodge sphinxarg mocking failure.
            flags = so3g.proj.RangesMatrix.zeros(
                shape=(tod.dets.count, tod.samps.count))

        # Compute fraction of samples
        weight = np.mean(flags.get_stats()['samples']) / flags.shape[1]
        logger.info(f'Total mask weight is {weight*100:.2}%.')

        # Wrap result into AxisManager for HDF5 off-load.
        aman = core.AxisManager(tod.dets, tod.samps)
        aman.wrap('source_flags', flags, [(0, 'dets'), (1, 'samps')])

        # Get file + dataset from policy.
        policy = util.ArchivePolicy.from_params(config['archive']['policy'])
        dest_file, dest_dataset = policy.get_dest(obs_id)
        aman.save(dest_file, dest_dataset, overwrite=True)

        # Update the index.
        db_data = {'obs:obs_id': obs_id,
                   'dataset': dest_dataset,
                   f'dets:{group_by}': this_index}
        db.add_entry(db_data, dest_file, replace=True)

    # Return something?
    return tod, aman

if __name__ == '__main__':
    util.main_launcher(main, get_parser)



import os
import yaml
import time
import logging
from typing import Optional, Union, Callable
import numpy as np
import argparse
import traceback
from typing import Optional
from sotodlib.utils.procs_pool import get_exec_env
import h5py
import copy
from sotodlib.coords import demod as demod_mm
from sotodlib.hwp import hwp_angle_model
from sotodlib import core
import sotodlib.site_pipeline.util as sp_util
from sotodlib.preprocess import preprocess_util as pp_util
from sotodlib.preprocess import _Preprocess, Pipeline, processes

logger = sp_util.init_logger("preprocess")

def dummy_preproc(obs_id, group_list, logger,
                  configs, overwrite, run_parallel):
    """
    Dummy function that can be put in place of preprocess_tod in the
    main function for testing issues in the processpoolexecutor
    (multiprocessing).
    """
    error = None
    outputs = []
    context = core.Context(configs["context_file"])
    group_by, groups, error = pp_util.get_groups(obs_id, configs, context)
    pipe = Pipeline(configs["process_pipe"], plot_dir=configs["plot_dir"], logger=logger)
    for group in groups:
        logger.info(f"Beginning run for {obs_id}:{group}")
        dets = {gb:gg for gb, gg in zip(group_by, group)}
        proc_aman = core.AxisManager(core.LabelAxis('dets', ['det%i' % i for i in range(3)]),
                                     core.OffsetAxis('samps', 1000))
        proc_aman.wrap_new('signal', ('dets', 'samps'), dtype='float32')
        proc_aman.wrap_new('timestamps', ('samps',))[:] = (np.arange(proc_aman.samps.count) / 200)

        outputs_grp = pp_util.save_group(obs_id, configs, dets, context, subdir='temp')
        logger.info(f"Saving data to {outputs_grp['temp_file']}:{outputs_grp['db_data']['dataset']}")
        proc_aman.save(outputs_grp['temp_file'], outputs_grp['db_data']['dataset'], overwrite)

        if run_parallel:
            outputs.append(outputs_grp)

    if run_parallel:
        return error, outputs


def preprocess_tod(obs_id,
                   configs,
                   verbosity=0,
                   group_list=None,
                   overwrite=False,
                   run_parallel=False):
    """Meant to be run as part of a batched script, this function calls the
    preprocessing pipeline a specific Observation ID and saves the results in
    the ManifestDb specified in the configs.

    Arguments
    ----------
    obs_id: string or ResultSet entry
        obs_id or obs entry that is passed to context.get_obs
    configs: string or dictionary
        config file or loaded config directory
    group_list: None or list
        list of groups to run if you only want to run a partial update
    overwrite: bool
        if True, overwrite existing entries in ManifestDb
    verbosity: log level
        0 = error, 1 = warn, 2 = info, 3 = debug
    run_parallel: Bool
        If true preprocess_tod is called in a parallel process which returns
        dB info and errors and does no sqlite writing inside the function.
    """

    outputs = []
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))

    context = core.Context(configs["context_file"])
    group_by, groups, error = pp_util.get_groups(obs_id, configs, context)

    if error is not None:
        if run_parallel:
            return error[0], [None, None]
        else:
            return

    all_groups = groups.copy()
    for g in all_groups:
        if group_list is not None:
            if g not in group_list:
                groups.remove(g)
                continue
        if 'wafer.bandpass' in group_by:
            if 'NC' in g:
                groups.remove(g)
                continue
        try:
            meta = context.get_meta(obs_id, dets = {gb:gg for gb, gg in zip(group_by, g)})
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {g}\n{errmsg}\n{tb}")
            groups.remove(g)
            continue

        if meta.dets.count == 0:
            groups.remove(g)

    if len(groups) == 0:
        logger.warning(f"group_list:{group_list} contains no overlap with "
                       f"groups in observation: {obs_id}:{all_groups}. "
                       f"No analysis to run.")
        error = 'no_group_overlap'
        if run_parallel:
            return error, [None, None]
        else:
            return

    if not(run_parallel):
        db = pp_util.get_preprocess_db(configs, group_by)

    pipe = Pipeline(configs["process_pipe"], plot_dir=configs["plot_dir"], logger=logger)

    if configs.get("lmsi_config", None) is not None:
        make_lmsi = True
    else:
        make_lmsi = False

    n_fail = 0
    for group in groups:
        logger.info(f"Beginning run for {obs_id}:{group}")
        dets = {gb:gg for gb, gg in zip(group_by, group)}
        try:
            aman = context.get_obs(obs_id, dets=dets)
            tags = np.array(context.obsdb.get(aman.obs_info.obs_id, tags=True)['tags'])
            aman.wrap('tags', tags)
            proc_aman, success = pipe.run(aman)

            if make_lmsi:
                new_plots = os.path.join(configs["plot_dir"],
                                         f'{str(aman.timestamps[0])[:5]}',
                                         aman.obs_info.obs_id)
        except Exception as e:
            #error = f'{obs_id} {group}'
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {group}\n{errmsg}\n{tb}")
            # return error, None, [errmsg, tb]
            # need a better way to log if just one group fails.
            n_fail += 1
            continue
        if success != 'end':
            # If a single group fails we don't log anywhere just mis an entry in the db.
            logger.info(f"ERROR: {obs_id} {group}\nFailed at step {success}")
            n_fail += 1
            continue

        outputs_grp = pp_util.save_group(obs_id, configs, dets, context, subdir='temp')
        logger.info(f"Saving data to {outputs_grp['temp_file']}:{outputs_grp['db_data']['dataset']}")
        proc_aman.save(outputs_grp['temp_file'], outputs_grp['db_data']['dataset'], overwrite)

        if run_parallel:
            outputs.append(outputs_grp)
        else:
            logger.info(f"Saving to database under {outputs_grp['db_data']}")
            if len(db.inspect(outputs_grp['db_data'])) == 0:
                h5_path = os.path.relpath(outputs_grp['temp_file'],
                        start=os.path.dirname(configs['archive']['index']))
                db.add_entry(outputs_grp['db_data'], h5_path)

    if make_lmsi:
        from pathlib import Path
        import lmsi.core as lmsi

        if os.path.exists(new_plots):
            lmsi.core([Path(x.name) for x in Path(new_plots).glob("*.png")],
                      Path(configs["lmsi_config"]),
                      Path(os.path.join(new_plots, 'index.html')))

    if run_parallel:
        if n_fail == len(groups):
            # If no groups make it to the end of the processing return error.
            logger.info(f'ERROR: all groups failed for {obs_id}')
            error = 'all_fail'
            return error, [obs_id, 'all groups']
        else:
            logger.info('Returning data to futures')
            error = None
            return error, outputs


def load_preprocess_tod_sim(obs_id, sim_map,
                            configs="preprocess_configs.yaml",
                            context=None, dets=None,
                            meta=None, modulated=True, logger=logger):
    """Loads the saved information from the preprocessing pipeline and runs the
    processing section of the pipeline on simulated data

    Assumes preprocess_tod has already been run on the requested observation.

    Arguments
    ----------
    obs_id: multiple
        passed to ``context.get_obs`` to load AxisManager, see Notes for
        `context.get_obs`
    sim_map: pixell.enmap.ndmap
        signal map containing (T, Q, U) fields
    configs: string or dictionary
        config file or loaded config directory
    dets: dict
        dets to restrict on from info in det_info. See context.get_meta.
    meta: AxisManager
        Contains supporting metadata to use for loading.
        Can be pre-restricted in any way. See context.get_meta.
    modulated: bool
        If True, apply the HWP angle model and scan the simulation
        into a modulated signal.
        If False, scan the simulation into demodulated timestreams.
    """
    configs, context = pp_util.get_preprocess_context(configs, context)
    if dets is not None:
        meta.restrict("dets", dets)
    det_vals = pp_util.load_preprocess_det_select(
        obs_id, configs=configs, context=context, logger=logger
    )
    meta.restrict("dets", [d for d in meta.dets.vals if d in det_vals])

    if meta.dets.count == 0:
        logger.info(f"No detectors left after cuts in obs {obs_id}")
        return None
    else:
        pipe = Pipeline(configs["process_pipe"], logger=logger)
        aman = context.get_obs(meta, no_signal=True)
        if modulated:
            # Apply the HWP angle model here
            # WARNING : should be turned off in the config file
            # to filter simulations
            aman = hwp_angle_model.apply_hwp_angle_model(aman)
            aman.move("signal", None)
        demod_mm.from_map(aman, sim_map, wrap=True, modulated=modulated)
        pipe.run(aman, aman.preprocess, sim=True)
        return aman


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('configs', help="Preprocessing Configuration File")
    parser.add_argument(
        '--query',
        help="Query to pass to the observation list. Use \\'string\\' to "
             "pass in strings within the query.",
        type=str
    )
    parser.add_argument(
        '--obs-id',
        help="obs-id of particular observation if we want to run on just one"
    )
    parser.add_argument(
        '--overwrite',
        help="If true, overwrites existing entries in the database",
        action='store_true',
    )
    parser.add_argument(
        '--min-ctime',
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max-ctime',
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--update-delay',
        help="Number of days (unit is days) in the past to start observation list.",
        type=int
    )
    parser.add_argument(
        '--tags',
        help="Observation tags. Ex: --tags 'jupiter' 'setting'",
        nargs='*',
        type=str
    )
    parser.add_argument(
        '--planet-obs',
        help="If true, takes all planet tags as logical OR and adjusts related configs",
        action='store_true',
    )
    parser.add_argument(
        '--verbosity',
        help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
        default=2,
        type=int
    )
    parser.add_argument(
        '--nproc',
        help="Number of parallel processes to run on.",
        type=int,
        default=4
    )
    parser.add_argument(
        '--raise-error',
        help="Raise an error upon completion if any obsids or groups fail.",
        type=bool,
        default=False
    )
    return parser


def _main(executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
          as_completed_callable: Callable,
          configs: str,
          query: Optional[str] = None,
          obs_id: Optional[str] = None,
          overwrite: bool = False,
          min_ctime: Optional[int] = None,
          max_ctime: Optional[int] = None,
          update_delay: Optional[int] = None,
          tags: Optional[str] = None,
          planet_obs: bool = False,
          verbosity: Optional[int] = None,
          nproc: Optional[int] = 4,
          raise_error: Optional[bool] = False):

    configs, context = pp_util.get_preprocess_context(configs)
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)

    errlog = os.path.join(os.path.dirname(configs['archive']['index']),
                          'errlog.txt')

    obs_list = sp_util.get_obslist(context, query=query, obs_id=obs_id, min_ctime=min_ctime,
                                   max_ctime=max_ctime, update_delay=update_delay, tags=tags,
                                   planet_obs=planet_obs)

    if len(obs_list)==0:
        logger.warning(f"No observations returned from query: {query}")

    # clean up lingering files from previous incomplete runs
    policy_dir = os.path.join(os.path.dirname(configs['archive']['policy']['filename']), 'temp')
    for obs in obs_list:
        obs_id = obs['obs_id']
        pp_util.cleanup_obs(obs_id, policy_dir, errlog, configs, context,
                            subdir='temp', remove=overwrite)

    # remove datasets from final archive file not found in db
    pp_util.cleanup_archive(configs, logger)

    run_list = []

    if overwrite or not os.path.exists(configs['archive']['index']):
        #run on all if database doesn't exist
        run_list = [ (o,None) for o in obs_list]
        group_by = np.atleast_1d(configs['subobs'].get('use', 'detset'))
    else:
        db = core.metadata.ManifestDb(configs['archive']['index'])

        for obs in obs_list:
            x = db.inspect({'obs:obs_id': obs["obs_id"]})
            if x is None or len(x) == 0:
                run_list.append( (obs, None) )
            else:
                group_by, groups, _ = pp_util.get_groups(obs["obs_id"], configs, context)
                if len(x) != len(groups):
                    [groups.remove([a[f'dets:{gb}'] for gb in group_by]) for a in x]
                    run_list.append( (obs, groups) )

    logger.info(f'Run list created with {len(run_list)} obsids')

    n_fail = 0

    # Run write_block obs-ids in parallel at once then write all to the sqlite db.
    futures = [executor.submit(preprocess_tod, obs_id=r[0]['obs_id'],
                    group_list=r[1], verbosity=verbosity,
                    configs=configs,
                    overwrite=overwrite, run_parallel=True) for r in run_list]
    for future in as_completed_callable(futures):
        logger.info('New future as_completed result')
        try:
            err, db_datasets = future.result()
            if err is not None:
                n_fail += 1
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: future.result()\n{errmsg}\n{tb}")
            f = open(errlog, 'a')
            f.write(f'\n{time.time()}, future.result() error\n{errmsg}\n{tb}\n')
            f.close()
            n_fail+=1
            continue
        futures.remove(future)

        if db_datasets:
            if err is None:
                logger.info(f'Processing future result db_dataset: {db_datasets}')
                for db_dataset in db_datasets:
                    pp_util.cleanup_mandb(err, db_dataset, configs, logger)
            else:
                pp_util.cleanup_mandb(err, db_datasets, configs, logger)

    if raise_error and n_fail > 0:
        raise RuntimeError(f"preprocess_tod: {n_fail}/{len(run_list)} obs_ids failed")

def main(configs: str,
         query: Optional[str] = None,
         obs_id: Optional[str] = None,
         overwrite: bool = False,
         min_ctime: Optional[int] = None,
         max_ctime: Optional[int] = None,
         update_delay: Optional[int] = None,
         tags: Optional[str] = None,
         planet_obs: bool = False,
         verbosity: Optional[int] = None,
         nproc: Optional[int] = 4,
         raise_error: Optional[bool] = False):

    rank, executor, as_completed_callable = get_exec_env(nproc)
    if rank == 0:
        _main(executor=executor,
              as_completed_callable=as_completed_callable,
              configs=configs,
              query=query,
              obs_id=obs_id,
              overwrite=overwrite,
              min_ctime=min_ctime,
              max_ctime=max_ctime,
              update_delay=update_delay,
              tags=tags,
              planet_obs=planet_obs,
              verbosity=verbosity,
              nproc=nproc,
              raise_error=raise_error)

if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)


# Copyright (c) 2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Check functionality of Splits.

"""
import os
import unittest

import astropy.units as u
import numpy as np

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import toast
    import toast.ops
    from toast.observation import default_values as defaults

    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    toast_available = True
except ImportError as e:
    toast_available = False

from ._helpers import (
    calibration_schedule,
    create_outdir,
    close_data_and_comm,
    simulation_test_data,
)

from sotodlib.toast.ops import pos_to_chi


class SplitTest(unittest.TestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        if not toast_available:
            print("TOAST cannot be imported- skipping unit tests", flush=True)
            return
        world, procs, rank = toast.get_world()
        self.outdir = create_outdir(fixture_name, mpicomm=world)

    def test_split_maps(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="SAT1",
            wafer_slot="w28",
            bands="SAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=16,
            cal_schedule=False,
        )

        pointing_radec = toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            quats="quats_radec",
            boresight=defaults.boresight_radec,
        )
        pointing_azel = toast.ops.PointingDetectorSimple(
            name="det_pointing_azel",
            quats="quats_azel",
            boresight=defaults.boresight_azel,
        )

        noise_model = toast.ops.DefaultNoiseModel(
            name="default_model", noise_model="noise_model"
        )
        noise_model.apply(data)

        el_model = toast.ops.ElevationNoise(
            name="elevation_model",
            out_model="noise_model",
            detector_pointing=pointing_azel,
        )
        el_model.apply(data)

        weights = toast.ops.StokesWeights(
            name="weights_radec",
            weights="weights_radec",
            mode="IQU",
            detector_pointing=pointing_radec,
        )

        pixels = toast.ops.PixelsHealpix(
            name="pixels_radec",
            detector_pointing=pointing_radec,
            nside=128,
        )

        binner = toast.ops.BinMap(
            name="binner",
            pixel_pointing=pixels,
            stokes_weights=weights,
        )

        mapmaker = toast.ops.MapMaker(
            name="mapmaker",
            binning=binner,
            map_binning=binner,
        )

        # Test that we can instantiate all the splits
        splits = so_ops.Splits(
            name="splits",
            splits=[
                "all",
                "left_going",
                "right_going",
                "outer_detectors",
                "inner_detectors",
                "polA",
                "polB",
            ],
            mapmaker=mapmaker,
            output_dir=self.outdir,
        )

        splits.apply(data)

        close_data_and_comm(data)


if __name__ == '__main__':
    unittest.main()



"""
Script to make hwp angle metadata.
This script will run:
1. automatically if data are acquired as preliminary version.
2. New versions of the metadata can be created as the analysis evolves.
"""
import os
import time
import argparse
import logging
import yaml
import traceback
import sqlite3
from typing import Optional, Union, Callable

from sotodlib import core
from sotodlib.hwp.g3thwp import G3tHWP
from sotodlib.site_pipeline import util
from sotodlib.utils.procs_pool import get_exec_env

max_trial = 5
wait_time = 5

# encodings for flacarray compression
encodings = {
    'hwp_angle': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver1_1': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver2_1': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver3_1': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver1_2': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver2_2': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_angle_ver3_2': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_rate_1': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'hwp_rate_2': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 1.0e-8}},
    'timestamps': {'type': 'flacarray', 'args': {'level': 5, 'quanta': 5.0e-5}},
    'quad_1': {'type': 'flacarray'},
    'quad_2': {'type': 'flacarray'},
}

logger = util.init_logger('make_hwp_solutions', 'make-hwp-solutions: ')


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument(
        'context',
        help="Path to context yaml file to define observation\
        for which to generate hwp angle.")
    parser.add_argument(
        'HWPconfig',
        help="Path to HWP configuration yaml file.")
    parser.add_argument(
        '-o', '--solution-output-dir', action='store', default=None, type=str,
        help='output directory of solution metadata, \
        overwrite config solution_output_dir')
    parser.add_argument(
        '--encoder-output-dir', action='store', default=None, type=str,
        help='output directory of encoder metadata, \
        overwrite config encoder_output_dir')
    parser.add_argument(
        '--verbose', default=2, type=int,
        help="increase output verbosity. \
        0: Error, 1: Warning, 2: Info(default), 3: Debug")
    parser.add_argument(
        '--overwrite', action='store_true',
        help="If true, overwrites existing entries in the database",
    )
    parser.add_argument(
        '--query', type=str,
        help="Query to pass to the observation list. Use \\'string\\' to "
             "pass in strings within the query.",
    )
    parser.add_argument(
        '--min-ctime',
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max-ctime',
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--obs-id', type=str, nargs='+',
        help="List of obs-ids of particular observations that you want to run",
    )
    parser.add_argument(
        '--off-site', action='store_true',
        help="This must be true when you run this off site. "
             "If true, this does not try to load L2 data"
    )
    parser.add_argument(
        '--nprocs', default=1, type=int,
        help="number of processes to use. We use nprocs=1"
             " at site computing"
    )
    return parser


def make_db(db_filename):
    """
    Make sqlite database
    Get file + dataset from policy.
    policy = util.ArchivePolicy.from_params(config['archive']['policy'])
    dest_file, dest_dataset = policy.get_dest(obs_id)
    Use 'output_dir' argument for now
    """
    if os.path.exists(db_filename):
        logger.info(f"Mapping {db_filename} for the "
                    "archive index.")
        db = core.metadata.ManifestDb(db_filename)
    else:
        logger.info(f"Creating {db_filename} for the "
                    "archive index.")
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(
            db_filename,
            scheme=scheme
        )
    return db


def save(aman, db, h5_filename, output_dir, obs_id, overwrite, compression, encodings=None):
    """
    Save HDF5 file and add entry to sqlite database
    Return True if save is succeeded, otherwise return False
    """
    aman_saved = False
    db_saved = False

    for i in range(1, max_trial + 1):
        try:
            aman.save(os.path.join(output_dir, h5_filename), obs_id, overwrite,
                      compression, encodings=encodings)
            logger.info("Saved aman")
            aman_saved = True
            break
        except BlockingIOError:
            logger.warning("Cannot save aman because HDF5 is temporary locked, "
                           f"try again in {wait_time} seconds, trial {i}/{max_trial}")
            time.sleep(wait_time)
        except Exception as e:
            logger.error(f"Exception '{e}' thrown while saving aman")
            print(traceback.format_exc())
            break
    if not aman_saved:
        logger.error("Cannot save aman, give up.")
        return False

    for i in range(1, max_trial + 1):
        try:
            db.add_entry(
                {'obs:obs_id': obs_id, 'dataset': obs_id}, filename=h5_filename, replace=overwrite,
            )
            logger.info("Added entry to db")
            db_saved = True
            break
        except sqlite3.OperationalError:
            logger.warning(f"Cannot save db, try again in {wait_time} seconds, trial {i}/{max_trial}")
            time.sleep(wait_time)
        except Exception as e:
            logger.error(f"Exception '{e}' thrown while saving aman")
            print(traceback.format_exc())
            break
    if not db_saved:
        logger.error("Cannot save db, give up.")
        return False

    return True


def get_solution(obs_id, encoder_completed, context, HWPconfig,
                 encoder_output_dir, off_site=False):
    """ calculate hwp angle solution.
    Load encoder L3 data if it exists.
    If this is run off site, load L2 data, otherwise skip it.

    Args
        obs_id: str
        encoder_completed:
            list of obs_id which has L3 encoder data

    Return
        obs_id: str
        aman_encoder: AxisManager
            axis manager of hwp encoder
        aman_solution: AxisManager
            axis manager of hwp angle solution
    """
    logger.info(f"Calculating Angles for {obs_id}")
    # split h5 file by first 5 digits of unixtime
    unix = obs_id.split('_')[1][:5]
    h5_encoder = os.path.join(encoder_output_dir, f'hwp_encoder_{unix}.h5')
    ctx = core.Context(context)
    g3thwp = G3tHWP(HWPconfig)

    # make angle solutions
    tod = ctx.get_obs(obs_id, dets=[])
    # Load L3 data if it exists
    if obs_id in encoder_completed:
        g3thwp.set_data(tod, h5_filename=h5_encoder)
        aman_encoder = None
    # Do nothing if L3 data does not exists off-site
    elif off_site:
        logger.info(f"Skip {obs_id} beause L3 data do not exist off site")
        return obs_id, None, None
    # Load L2 data at site
    else:
        aman_encoder = g3thwp.set_data(tod)

    aman_solution = g3thwp.make_solution(tod)
    return obs_id, aman_encoder, aman_solution


def _main(
    executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
    as_completed_callable: Callable,
    context: str,
    HWPconfig: str,
    solution_output_dir: Optional[str] = None,
    encoder_output_dir: Optional[str] = None,
    verbose: Optional[int] = 2,
    overwrite: Optional[bool] = False,
    query: Optional[str] = None,
    min_ctime: Optional[float] = None,
    max_ctime: Optional[float] = None,
    obs_id: Optional[str] = None,
    off_site: Optional[bool] = False,
):

    logger.info(f"Using context {context} and HWPconfig {HWPconfig}")

    configs = yaml.safe_load(open(HWPconfig, "r"))
    logger.info("Starting make_hwp_solutions")

    # Specify output directory
    if solution_output_dir is None:
        solution_output_dir = configs["solution_output_dir"]
    if encoder_output_dir is None:
        encoder_output_dir = configs["encoder_output_dir"]

    if not os.path.exists(solution_output_dir):
        logger.info(f"Making output directory {solution_output_dir}")
        os.mkdir(solution_output_dir)
    if not os.path.exists(encoder_output_dir):
        logger.info(f"Making output directory {encoder_output_dir}")
        os.mkdir(encoder_output_dir)

    # Set verbose
    if verbose == 0:
        logger.setLevel(logging.ERROR)
    elif verbose == 1:
        logger.setLevel(logging.WARNING)
    elif verbose == 2:
        logger.setLevel(logging.INFO)
    elif verbose == 3:
        logger.setLevel(logging.DEBUG)

    db_solution = make_db(os.path.join(solution_output_dir, 'hwp_angle.sqlite'))
    db_encoder = make_db(os.path.join(encoder_output_dir, 'hwp_encoder.sqlite'))
    encoder_completed = db_encoder.get_entries(['dataset'])['dataset']

    # load observation data
    if obs_id is not None:
        tot_query = ' or '.join([f"(obs_id=='{o}')" for o in obs_id])
    else:
        tot_query = "and "
        if min_ctime is not None:
            tot_query += f"start_time>={min_ctime} and "
        if max_ctime is not None:
            tot_query += f"start_time<={max_ctime} and "
        if query is not None:
            tot_query += query + " and "
        tot_query = tot_query[4:-4]
        if tot_query == "":
            tot_query = "1"

    logger.debug(f"Sending query to obsdb: {tot_query}")
    ctx = core.Context(context)
    obs_list = ctx.obsdb.query(tot_query)
    obs_list = [obs['obs_id'] for obs in obs_list]
    run_list = []

    if len(obs_list) == 0:
        logger.warning(f"No observations returned from query: {query}")

    completed = db_solution.get_entries(['dataset'])['dataset']

    for obs_id in obs_list:
        if overwrite or obs_id not in completed:
            run_list.append(obs_id)

    # write solutions
    futures = []
    for obs_id in run_list:
        futures.append(executor.submit(get_solution, obs_id, encoder_completed,
                                       context, HWPconfig, encoder_output_dir,
                                       off_site))

    for future in as_completed_callable(futures):
        try:
            obs_id, aman_encoder, aman_solution = future.result()
            # split h5 file by first 5 digits of unixtime
            unix = obs_id.split('_')[1][:5]
            h5_encoder = f'hwp_encoder_{unix}.h5'
            h5_solution = f'hwp_angle_{unix}.h5'
            if aman_encoder is not None:
                logger.info(f"Saving hwp_encoder: {obs_id}")
                success = save(aman_encoder, db_encoder, h5_encoder,
                               encoder_output_dir,
                               obs_id, overwrite, 'gzip')
                if not success:
                    logger.error(f"Failed to save hwp_encoder: {obs_id}, skip saving hwp_angle")
                    continue
            if aman_solution is not None:
                logger.info(f"Saving hwp_angle: {obs_id}")
                success = save(aman_solution, db_solution, h5_solution,
                               solution_output_dir, obs_id, overwrite,
                               compression='gzip', encodings=encodings)
                if not success:
                    logger.error(f"Failed to save hwp_angle: {obs_id}")
        except Exception as e:
            logger.error(e)
        del aman_encoder, aman_solution
        futures.remove(future)
    return


def main(context: str,
         HWPconfig: str,
         solution_output_dir: Optional[str] = None,
         encoder_output_dir: Optional[str] = None,
         verbose: Optional[int] = 2,
         overwrite: Optional[bool] = False,
         query: Optional[str] = None,
         min_ctime: Optional[float] = None,
         max_ctime: Optional[float] = None,
         obs_id: Optional[str] = None,
         off_site: Optional[bool] = False,
         nprocs: Optional[int] = 1):

    rank, executor, as_completed_callable = get_exec_env(nprocs=nprocs)
    if rank == 0:
        _main(executor=executor,
              as_completed_callable=as_completed_callable,
              context=context,
              HWPconfig=HWPconfig,
              solution_output_dir=solution_output_dir,
              encoder_output_dir=encoder_output_dir,
              verbose=verbose,
              overwrite=overwrite,
              query=query,
              min_ctime=min_ctime,
              max_ctime=max_ctime,
              obs_id=obs_id,
              off_site=off_site)


if __name__ == '__main__':
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))



from matplotlib.axes import Axes
from matplotlib.ticker import Formatter, MultipleLocator
from matplotlib.projections import register_projection

import numpy as np
from pixell import enmap

__all__ = ['BeamMapAxes']


class _RescaleFormatter(Formatter):
    """Tick formatter that divides the tick positions by some unit, and
    rounds the result to some number of decimal places for display.

    """
    def __init__(self, unit=1., precision=0):
        self.unit = unit
        self.precision = precision
    def __call__(self, x, pos=None):
        label = x / self.unit
        if self.precision == 0:
            return str(int(np.round(label)))
        return str(np.round(label, self.precision))


class BeamMapAxes(Axes):
    """Axes sub-class adapted for displaying beam maps.  This extends
    standard Axes in two ways:

    - Arrays passed to imshow for rendering must in fact be
      enmap.ndarrays; the wcs will be inspected to set the image
      extent (in degrees).
    - The X and Y scales and axis labels will be automatically
      rescaled, depending on the current viewport, to show
      degrees/arcminutes/arcseconds as appropriate.

    Args:
      unit (str): Force the axis & label units to one of 'deg',
        'arcmin', 'arcsec'.  If this is not set, the units will adapt
        to the viewport.
      n_major (int or (int, int)): The maximum number of major ticks
        to permit on the x and y axes.

    Notes:

      This class isn't used directly but rather is registered in
      matplotlib's projection system under the name "so-beammap",
      which can then be used as a projection argument to pyplot.axes
      or pyplot.subplot.  E.g.::

        ax = plt.subplot(221, projection='so-beammap')

      You can also use the .factor classmethod to achieve the same thing::

        ax = plt.subplot(221, projection=BeamMapAxes.factory())

      This latter form allows for defaults to be overridden::

        ax = plt.subplot(221, projection=BeamMapAxes.factory(units='arcmin', max_n=12))

      You might also find it useful to set the same projection on a
      set of subplots; that's done like this:

          fig, axs = plt.subplots(1, 3, subplot_kwargs={'projection': BeamMapAxes()})

    """
    name = 'so-beammap'

    rules = [
        # min_extent, unitstr
        (    2,  'deg'  ),
        ( 2/60, 'arcmin'),
        (    0, 'arcsec')
    ]

    def __init__(self, fig, rect, unit=None, n_major=None, **kw):
        self._force_unit = unit
        try:
            n_major = tuple(n_major)
        except TypeError:
            n_major = (n_major, n_major)
        self.n_major = n_major
        super().__init__(fig, rect, **kw)

    def _set_lim_helper(self, axis, ex, n):
        if n is None:
            n = 8

        # Extent and upper bound tick spacing
        extent = abs(ex[1] - ex[0])
        step = extent / n

        label = self._force_unit
        if label is None:
            for min_size, label in self.rules:
                if extent > min_size:
                    break
        unit = {'deg': 1., 'arcmin': 1/60, 'arcsec': 1/3600}[label]

        if step < unit:
            # If ticks will be smaller than 1 unit, start at some
            # 10**-n.
            base = 10
            zooms = []
        else:
            # But for ticks that look to be integer multiples of a
            # unit; start at 60**n, and go up in nice factors of 60.
            base = 60
            zooms = [2, 2.5, 2, 1.5, 2, 2]

        step = unit * base**np.floor(np.log(step/unit) / np.log(base))
        while round(abs(extent) / step) > n:
            if not zooms:
                zooms = [2, 2.5, 2]  # 10s.
            step *= zooms.pop(0)

        precision = max(int(np.ceil(np.log10(unit / step))),0)
        axis.set_major_locator(MultipleLocator(step))
        axis.set_major_formatter(_RescaleFormatter(unit, precision))

        return label

    def set_xlim(self, *args, **kw):
        super().set_xlim(*args, **kw)
        label = self._set_lim_helper(self.xaxis, self.get_xlim(), self.n_major[0])
        self.set_xlabel(f'$\\xi$ ({label})')

    def set_ylim(self, *args, **kw):
        super().set_ylim(*args, **kw)
        label = self._set_lim_helper(self.yaxis, self.get_ylim(), self.n_major[1])
        self.set_ylabel(f'$\\eta$ ({label})')

    def imshow(self, *args, **kwargs):
        """Wrapper for Axes.imshow with the following adjustments:

        - The image you pass in must be a 2d enmap.ndarray, i.e. it
          should have a ``wcs`` attribute.
        - kwarg 'origin' must be 'lower' (or omitted)
        - kwarg 'extent' must be None (or omitted); the boundary of
          the map, in degrees, will be used to set the "extent".  (As
          a result, means that any viewport adjustments should be
          passed in degrees, rather than pixel index coords or
          whatever units are being shown on the axis labels.)
        - kwarg 'aspect' will be set to match the pixel aspect ratio,
          unless it is explicitly passed in.

        """
        assert(kwargs.pop('origin', 'lower') == 'lower')
        assert(kwargs.pop('extent', None) is None)

        # just make extent be degrees!
        img = args[0]
        assert(isinstance(img, enmap.ndmap))
        wcs = img.wcs.wcs
        if kwargs.get('aspect') is None:
            # H / W of the pixel
            kwargs['aspect'] = abs(wcs.cdelt[1] / wcs.cdelt[0])
        extent = []
        for n, ax, sgn in [(img.shape[1], 0, -1), (img.shape[0], 1, 1)]:
            lohi = ((np.array([0, n]) - 0.5) + 1 - wcs.crpix[ax]) * wcs.cdelt[ax] + wcs.crval[ax]
            extent.extend(list(lohi * sgn))
        return super().imshow(*args, origin='lower', extent=extent, **kwargs)

    @classmethod
    def factory(cls, **kw):
        class _AxesFactory:
            def __init__(self, **kw):
                self.kw = kw
            def _as_mpl_axes(self):
                return cls, self.kw
        return _AxesFactory(**kw)


# Register 'so-beammap' with matplotlib
register_projection(BeamMapAxes)



# Copyright (c) 2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check tod_ops routines.

"""

import unittest
import numpy as np

from sotodlib import core
from sotodlib.hwp import hwp

INPUT_COEFFS = np.asarray(
                [[14.4086668, 10.57294344, 11.18964606, 12.27585103,
                 11.27456133, 11.7841095, 12.87044498, 13.25598597,
                 12.56675035, 12.06382477, 14.72865531, 12.62258284,
                 11.43530194, 11.92264568, 14.71371932, 11.8379756],
                [10.32748604, 11.82000625, 14.96949369, 14.18303644,
                 13.45385166, 12.81443907, 13.65328508, 13.78361639,
                 11.3964662, 13.06204199, 10.35769728, 10.20522442,
                 14.79385459, 11.24546604, 10.27690037, 10.04351162]])


def make_fake_tod(ts=np.arange(0, 60, 1/200), modes=np.arange(8)+1,
                  mult_fact=12.5*np.ones(8), ndets=2, input_coeffs=None):
    hwp_angle = (2*np.pi*2*ts) % (2*np.pi)
    fake_signal = np.zeros((ndets, len(hwp_angle)))
    if input_coeffs is None:
        input_coeffs = np.zeros((ndets, 2*len(mult_fact)))
    for nd in range(ndets):
        for i, md in enumerate(modes):
            if input_coeffs is None:
                input_coeffs[nd, 2*i] = mult_fact[i]*np.random.rand()
                input_coeffs[nd, 2*i+1] = mult_fact[i]*np.random.rand()
            fake_signal[nd] += input_coeffs[nd, 2*i]*np.sin(md*hwp_angle)
            fake_signal[nd] += input_coeffs[nd, 2*i + 1]*np.cos(md*hwp_angle)

    dets = ['det%i' % i for i in range(ndets)]
    mode_names = []
    for mode in modes:
        mode_names.append(f'S{mode}')
        mode_names.append(f'C{mode}')

    tod = core.AxisManager(core.LabelAxis('dets', vals=dets),
                           core.OffsetAxis('samps', count=len(hwp_angle)),
                           core.LabelAxis('modes', vals=mode_names))
    tod.wrap('timestamps', ts, axis_map=[(0, 'samps')])
    tod.wrap('hwp_angle', hwp_angle, axis_map=[(0, 'samps')])
    tod.wrap('signal', np.atleast_2d(fake_signal),
             axis_map=[(0, 'dets'), (1, 'samps')])
    tod.wrap('input_coeffs', np.atleast_2d(input_coeffs),
             axis_map=[(0, 'dets'), (1, 'modes')])
    return tod

def get_coeff_metric(tod):
    nm = tod.modes.count//2
    nd = tod.dets.count
    outmetric_num = np.zeros((nd, nm))
    outmetric_denom = np.zeros((nd, nm))
    for i in range(nm):
        outmetric_num[:, i] = (tod.hwpss_stats.coeffs[:, 2*i] - \
                               tod.input_coeffs[: , 2*i])**2
        outmetric_num[:, i] += (tod.hwpss_stats.coeffs[:, 2*i+1] - \
                                tod.input_coeffs[: , 2*i+1])**2
        outmetric_denom[:, i] = (tod.input_coeffs[:, 2*i])**2
        outmetric_denom[:, i] += (tod.input_coeffs[:, 2*i+1])**2
    return np.max(100*(outmetric_num/outmetric_denom))



class HwpssTest(unittest.TestCase):
    "Test the HWPSS fitting functions"
    def test_linregbin(self):
        lr, bn = True, True
        tod = make_fake_tod(input_coeffs=INPUT_COEFFS)
        _ = hwp.get_hwpss(tod, lin_reg=lr, bin_signal=bn,
                           bins=200)
        ommax = get_coeff_metric(tod)
        self.assertTrue(ommax < 0.1)

    def test_linregnobin(self):
        lr, bn = True, False
        tod = make_fake_tod(input_coeffs=INPUT_COEFFS)
        _ = hwp.get_hwpss(tod, lin_reg=lr, bin_signal=bn)
        ommax = get_coeff_metric(tod)
        self.assertTrue(ommax < 0.1)

    def test_fitbin(self):
        lr, bn = False, True
        tod = make_fake_tod(input_coeffs=INPUT_COEFFS)
        _ = hwp.get_hwpss(tod, lin_reg=lr, bin_signal=bn, bins=200)
        ommax = get_coeff_metric(tod)
        self.assertTrue(ommax < 0.1)
        # Checks that returned covariance matrix from fit is finite.
        # When not using wrapper function in lambda function we get
        # infinite covariance matrix returned. So this is a warning
        # to not change that even though it looks wrong.
        self.assertFalse(False in np.isfinite(tod.hwpss_stats.covars))

    def test_fitnobin(self):
        lr, bn = False, False
        tod = make_fake_tod(input_coeffs=INPUT_COEFFS)
        with self.assertRaises(ValueError):
            _ = hwp.get_hwpss(tod, lin_reg=lr, bin_signal=bn, modes=[2, 4])

if __name__ == '__main__':
    unittest.main()
